# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains helper functions for tests.
"""
import operator
import os
import networkx.algorithms.isomorphism as iso


def make_into_set(iter_of_dict):
    """
    Convenience function that turns an iterator of dicts into a set of
    frozenset of the dict items.
    """
    return set(frozenset(dict_.items()) for dict_ in iter_of_dict)


def equal_graphs(g1, g2,
                 node_attrs=('resid', 'resname', 'atomname', 'chain', 'charge_group', 'atype'),
                 edge_attrs=()):
    """
    Parameters
    ----------
    g1: networkx.Graph
    g2: networkx.Graph
    node_attrs: collections.abc.Iterable or None
        Node attributes to consider. If `None`, the node attribute dicts must
        be equal.
    edge_attrs: collections.abc.Iterable or None
        Edge attributes to consider. If `None`, the edge attribute dicts must
        be equal.

    Returns
    -------
    bool
        True if `g1` and `g2` are isomorphic, False otherwise.
    """
    if node_attrs is None:
        node_equal = operator.eq
    else:
        node_equal = iso.categorical_node_match(node_attrs, [''] * len(node_attrs))
    if edge_attrs is None:
        edge_equal = operator.eq
    else:
        edge_equal = iso.categorical_node_match(edge_attrs, [''] * len(edge_attrs))
    matcher = iso.GraphMatcher(g1, g2, node_match=node_equal, edge_match=edge_equal)
    return matcher.is_isomorphic()


def find_in_path(names=('martinize2', 'martinize2.py')):
    """
    Finds and returns the location of one of `names` in PATH, and returns the
    first match.

    Parameters
    ----------
    names: collections.abc.Sequence
        Names to look for in PATH.

    Returns
    -------
    os.PathLike or None
    """
    for folder in os.getenv("PATH", '').split(os.pathsep):
        for name in names:
            fullpath = os.path.join(folder, name)
            if os.path.isfile(fullpath):
                return fullpath


# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict, OrderedDict, namedtuple
import copy
from functools import partial
import itertools

import networkx as nx
import numpy as np

from . import graph_utils
from . import geometry
from . import utils


Interaction = namedtuple('Interaction', 'atoms parameters meta')
DeleteInteraction = namedtuple('DeleteInteraction',
                               'atoms atom_attrs parameters meta')


class LinkPredicate:
    """
    Comparison criteria for node and molecule attributes in links.

    When comparing an attribute from a link to a corresponding attribute from
    a molecule or a molecule node, the default behavior is to use the equality
    as criterion for the correspondence. Some correspondence, however must be
    broader for the link to be usable. Such alternative criteria are defined
    as link predicates.

    If an attribute in a link is set to an instance of a predicate, then the
    correspondence is defined as the boolean result of the ``match`` method.

    This is the base class for such predicate. It must be subclassed, and
    subclasses must define a :meth:`match` method that takes a dictionary and
    a potential key from that dictionary as arguments.

    Parameters
    ----------
    value:
        The per-instance value that serve as reference. How this value is
        treated depends on the subclass.
    """
    def __init__(self, value):
        self.value = value

    def match(self, node, key):
        """
        Do the comparison with the reference value.

        Notes
        -----
        This function **must** be defined by the subclasses. This docstring
        describe the *expected* format of the method.

        Parameters
        ----------
        node: dict
            A dictionary of attributes in which to look up. This can be a
            node dictionary of a molecule ``meta`` attribute.
        key:
            A potential key from the ``node`` dictionary.

        Returns
        -------
        bool
        """
        raise NotImplementedError

    def __eq__(self, other):
        # Should maybe be:
        # return (isinstance(other, self.__class__) or isinstance(self, other.__class__))\
        #        and self.value == other.value
        return other.__class__ == self.__class__ and self.value == other.value

    def __repr__(self):
        return '<{} at {:x} value={}>'.format(self.__class__.__name__, id(self), self.value)


class Choice(LinkPredicate):
    """
    Test if an attribute is defined and in a predefined list.

    Parameters
    ----------
    value: list
        The list of value in which to look for the attribute.
    """
    def match(self, node, key):
        """
        Apply the comparison.
        """
        return node.get(key) in self.value


class NotDefinedOrNot(LinkPredicate):
    """
    Test if an attribute is not the reference value.

    This test passes if the attribute is not defined, if it is set to ``None``,
    or if its value is different from the reference.

    Notes
    -----
    If the reference is set to ``None``, then the test does not pass if the
    attribute is explicitly set to ``None``. It still passes if the attribute
    is not defined.

    Parameters
    ----------
    value:
        The value the attribute is tested not to be.
    """
    def match(self, node, key):
        """
        Apply the comparison.
        """
        return key not in node or node[key] != self.value


class LinkParameterEffector:
    """
    Rule to calculate an interaction parameter in a link.

    This class allows to store dynamic parameters in link interactions. The
    value of the parameter can be computed from the graph using the node keys
    given when creating the instance.

    An instance of this class is first initialized with a list of node keys
    from the link in which it is defined. The instance is latter called
    like a function, and takes as arguments a molecule and a match dictionary
    linking the link nodes with the molecule ones. The format of the dictionary
    is expected to be ``{link key: molecule key}``.

    An instance can also have a format defined. If defined, that format will be
    applied to the value computed by the :meth:`_apply` method causing the
    output to be a string. The format is given as a 'format_spec' from the
    python format string syntax. This format spec corresponds to what follows
    the column the column in string templates. For instance, formating a
    floating number to have 2 decimal places will be obtained by setting format
    to `.2f`. If no format is defined, then the calculated value is not
    modified.

    This is a base class; it needs to be subclassed. A subclass must define an
    :meth:`_apply` method that takes a molecule and a list of node keys from
    that molecule as arguments. This method is not called directly by the user,
    instead it is called by the :meth:`__call__` method when the user calls the
    instance as a function. A subclass can also set the :attr:`n_keys_asked`
    class attribute to the number of required keys. If the
    attribute is set, then the number of keys provided when initializing a new
    instance will be validated against that number; else, the user can pass an
    arbitrary number of keys without validation.

    .. automethod:: __call__
    .. automethod:: _apply
    """
    n_keys_asked = None
    """Class attribute describing the number of keys required."""

    def __init__(self, keys, format_spec=None):
        """
        Parameters
        ----------
        keys: list
            A list of node keys from the link. If the :attr:`n_keys_asked`
            class argument is set, the number of keys must correspond to the
            value of the attribute.
        format_spec: str
            Format specification.

        Raises
        ------
        ValueError
            Raised if the :attr:`n_keys_asked` class attribute is set and the
            number of keys does not correspond.
        """
        self.keys = keys
        if self.n_keys_asked is not None and len(self.keys) != self.n_keys_asked:
            raise ValueError(
                'Unexpected number of keys provided in {}: '
                '{} were expected, but {} were provided.'
                .format(self.__class__.__name__, self.n_keys_asked, len(keys))
            )
        self.format = format_spec

    def __eq__(self, other):
        return (self.__class__ == other.__class__
                and self.keys == other.keys
                and self.format == other.format)

    def __call__(self, molecule, match):
        """
        Parameters
        ----------
        molecule: Molecule
            The molecule from which to calculate the parameter value.
        match: dict
            The correspondence between the nodes from the link (keys), and the
            nodes from the molecule (values).

        Returns
        -------
        float:
            The calculated parameter value, formatted if required.
        """
        keys = [match[key] for key in self.keys]
        result = self._apply(molecule, keys)
        if self.format is not None:
            result = '{value:{format}}'.format(value=result, format=self.format)
        return result

    def _apply(self, molecule, keys):
        """
        Calculate the parameter value from the molecule.

        Notes
        -----
        This method **must** be defined in a subclass.

        Parameters
        ----------
        molecule: Molecule
            The molecule from which to compute the parameter value.
        keys: list
            A list of keys to use from the molecule.

        Returns
        -------
        float:
            The value for the parameter.
        """
        msg = 'The method need to be implemented by the children class.'
        raise NotImplementedError(msg)


class ParamDistance(LinkParameterEffector):
    """
    Calculate the distance between a pair of nodes.
    """
    n_keys_asked = 2

    def _apply(self, molecule, keys):
        # This will raise a ValueError if an atom is missing, or if an
        # atom does not have position.
        positions = np.stack([molecule.nodes[key]['position'] for key in keys])
        # We assume there are two rows; which we can since we checked earlier
        # that exactly two atom keys were passed.
        distance = np.sqrt(np.sum(np.diff(positions, axis=0)**2))
        return distance


class ParamAngle(LinkParameterEffector):
    """
    Calculate the angle in degrees between three consecutive nodes.
    """
    n_keys_asked = 3

    @staticmethod
    def _apply(molecule, keys):
        # This will raise a ValueError if an atom is missing, or if an
        # atom does not have position.
        positions = np.stack([molecule.nodes[key]['position'] for key in keys])
        vector_ba = positions[0, :] - positions[1, :]
        vector_bc = positions[2, :] - positions[1, :]
        angle = geometry.angle(vector_ba, vector_bc)
        return np.degrees(angle)


class ParamDihedral(LinkParameterEffector):
    """
    Calculate the dihedral angle in degrees defined by four nodes.
    """
    n_keys_asked = 4

    @staticmethod
    def _apply(molecule, keys):
        # This will raise a ValueError if an atom is missing, or if an
        # atom does not have position.
        positions = np.stack([molecule.nodes[key]['position'] for key in keys])
        angle = geometry.dihedral(positions)
        return np.degrees(angle)


class ParamDihedralPhase(LinkParameterEffector):
    """
    Calculate the dihedral angle in degrees defined by four nodes shifted by
    -180 degrees.
    """
    n_keys_asked = 4

    @staticmethod
    def _apply(molecule, keys):
        # This will raise a ValueError if an atom is missing, or if an
        # atom does not have position.
        positions = np.stack([molecule.nodes[key]['position'] for key in keys])
        angle = geometry.dihedral_phase(positions)
        return np.degrees(angle)


class Molecule(nx.Graph):
    """
    Represents a molecule as per a specific force field. Consists of atoms
    (nodes), bonds (edges) and interactions such as angle potentials.

    Two molecules are equal if:

    * the exclusion distance (nrexcl) are equal
    * the force fields are equal (but may be different instances)
    * the nodes are equal and in the same order
    * the edges are equal (but order is not accounted for)
    * the interactions are the same and in the same order within an interaction
      type

    When comparing molecules, the order of the nodes is considered as it
    determines in what order atoms will be written in the output. Same goes for
    the interactions within an interaction type. The order of edges is not
    guaranteed anywhere in the code, and they are not writen in the output.
    """
    # As the particles are stored as nodes, we want the nodes to stay
    # ordered.
    node_dict_factory = OrderedDict

    def __init__(self, *args, **kwargs):
        self.meta = kwargs.pop('meta', {})
        self._force_field = kwargs.pop('force_field', None)
        self.nrexcl = kwargs.pop('nrexcl', None)
        super().__init__(*args, **kwargs)
        self.interactions = defaultdict(list)

    def __eq__(self, other):
        return (
            self.nrexcl == other.nrexcl and
            self._force_field == other._force_field and
            self.same_nodes(other) and
            self.same_edges(other) and
            self.same_interactions(other)
        )

    @staticmethod
    def sort_interactions(all_interactions):
        """
        Returns keys in interactions sorted by (number_of_atoms, name). Keys
        with no interactions are skipped.
        """
        sort_keys = {}
        for interaction_type, interactions in all_interactions.items():
            if not interactions:
                continue
            sort_keys[interaction_type] = len(interactions[0].atoms), interaction_type
        return sorted(sort_keys, key=lambda k: sort_keys[k])

    def __str__(self):
        moltype = self.meta.get('moltype', 'molecule')

        interaction_count = OrderedDict()
        # Make sure atoms and edges get sorted first.
        interaction_count['atoms'] = len(self.nodes)
        if len(self.interactions.get('bonds', [])) != len(self.edges):
            interaction_count['edges'] = len(self.edges)

        for itype in self.sort_interactions(self.interactions):
            interaction_count[itype] = len(self.interactions[itype])

        # interaction_count will always contain at least 'atoms'.
        last_item = interaction_count.popitem(last=True)
        out = "{} with ".format(moltype)
        out += ', '.join('{} {}'.format(number, itype)
                         for itype, number in interaction_count.items())
        if interaction_count:
            out += ', and '
        out += '{} {}'.format(last_item[1],
                              last_item[0])
        return out

    @property
    def force_field(self):
        """
        The force field the molecule is described for.

        The force field is assumed to be consistent for all the molecules of
        a system. While it is possible to reassign attribute
        `Molecule._force_field`, it is recommended to assign the force
        field at the system level as reassigning :attr:`~vermouth.system.System.force_field`
        will propagate the change to all the molecules in that system.
        """
        return self._force_field

    @property
    def atoms(self):
        """
        All atoms in this molecule. Alias for `nodes`.

        See Also
        --------
        :meth:`networkx.Graph.nodes`
        """
        # TODO: should just be an alias for nodes. If you need the attributes,
        #       do g.nodes(data=<attr>) or g.nodes(data=True)
        for node in self.nodes():
            node_attr = self.nodes[node]
            yield node, node_attr

    def copy(self):
        """
        Creates a copy of the molecule.

        Returns
        -------
        Molecule
        """
        new = self.subgraph(self.nodes)
        new.name = self.name
        return new

    def subgraph(self, nodes):
        """
        Creates a subgraph from the molecule.


        Returns
        -------
        Molecule
        """
        subgraph = self.__class__()
        subgraph.name = self.name
        subgraph.meta = copy.copy(self.meta)
        subgraph._force_field = self._force_field
        subgraph.nrexcl = self.nrexcl
        subgraph.name = self.name

        node_copies = [(node, copy.copy(self.nodes[node])) for node in nodes]
        subgraph.add_nodes_from(node_copies)

        nodes = set(nodes)

        #edges_to_add = [
        #    (node, node2)
        #    for node in nodes
        #    for node2 in set(self[node]) & nodes
        #]
        subgraph.add_edges_from(self.edges_between(nodes, nodes, data=True))

        for interaction_type, interactions in self.interactions.items():
            for interaction in interactions:
                if all(atom in nodes for atom in interaction.atoms):
                    subgraph.interactions[interaction_type].append(interaction)

        return subgraph

    def add_interaction(self, type_, atoms, parameters, meta=None):
        """
        Add an interaction of the specified type with the specified parameters
        involving the specified atoms.

        Parameters
        ----------
        type_: str
            The type of interaction, such as 'bonds' or 'angles'.
        atoms: collections.abc.Sequence
            The atoms that are involved in this interaction. Must be in this
            molecule
        parameters: collections.abc.Iterable
            The parameters for this interaction.
        meta: collections.abc.Mapping
            Metadata for this interaction, such as comments to be written to
            the output.

        Raises
        ------
        KeyError
            If one of the atoms is not in this molecule.
        """
        if meta is None:
            meta = {}
        for atom in atoms:
            if atom not in self:
                raise KeyError('Unknown atom {}'.format(atom))
        self.interactions[type_].append(
            Interaction(atoms=tuple(atoms), parameters=parameters, meta=meta)
        )

    def add_or_replace_interaction(self, type_, atoms, parameters, meta=None):
        """
        Adds a new interaction if it doesn't exists yet, and replaces it
        otherwise. Interactions are deemed the same if they're the same type,
        and they involve the same atoms, and their ``meta['version']`` is the
        same.

        Parameters
        ----------
        type_: str
            The type of interaction, such as 'bonds' or 'angles'.
        atoms: collections.abc.Sequence
            The atoms that are involved in this interaction. Must be in this
            molecule
        parameters: collections.abc.Iterable
            The parameters for this interaction.
        meta: collections.abc.Mapping
            Metadata for this interaction, such as comments to be written to
            the output.

        See Also
        --------
        :meth:`add_interaction`
        """
        if meta is None:
            meta = {}
        for idx, interaction in enumerate(self.interactions[type_]):
            if (interaction.atoms == tuple(atoms)
                    and interaction.meta.get('version', 0) == meta.get('version', 0)):
                new_interaction = Interaction(
                    atoms=tuple(atoms), parameters=parameters, meta=meta,
                )
                self.interactions[type_][idx] = new_interaction
                break
        else:  # no break
            self.add_interaction(type_, atoms, parameters, meta)

    def get_interaction(self, type_):
        """
        Returns all interactions of `type_`

        Parameters
        ----------
        type_: collections.abc.Hashable
            The type which interactions should be found.

        Returns
        -------
        list[Interaction]
            The interactions of the specified type.
        """
        return self.interactions[type_]

    def remove_interaction(self, type_, atoms, version=0):
        """
        Removes the specified interaction.

        Parameters
        ----------
        type_: str
            The type of interaction, such as 'bonds' or 'angles'.
        atoms: collections.abc.Sequence
            The atoms that are involved in this interaction.
        version: int
            Sometimes there can be multiple distinct interactions between the
            same group of atoms. This is reflected with their `version` meta
            attribute.

        Raises
        ------
        KeyError
            If the specified interaction could not be found
        """
        idx = 0
        for idx, interaction in enumerate(self.interactions[type_]):
            if interaction.atoms == atoms and interaction.meta.get('version', 0):
                break
        else:  # no break
            msg = ("Can't find interaction of type {} between atoms {} "
                   "and with version {}")
            raise KeyError(msg.format(type_, atoms, version))
        del self.interactions[type_][idx]

    def remove_matching_interaction(self, type_, template_interaction):
        """
        Removes any interactions that match the template.

        Parameters
        ----------
        type_: collections.abc.Hashable
            The type of interaction to look for.
        template_interaction: Interaction

        See Also
        --------
        :func:`interaction_match`
        """
        for idx, interaction in enumerate(self.interactions[type_]):
            if interaction_match(self, interaction, template_interaction):
                del self.interactions[type_][idx]
                break
        else:  # no break
            raise ValueError('Cannot find a matching interaction.')

    def find_atoms(self, **attrs):
        """
        Yields all indices of atoms that match `attrs`

        Parameters
        ----------
        **attrs: collections.abc.Mapping
            The attributes and their desired values.

        Yields
        ------
        collections.abc.Hashable
            All atom indices that match the specified `attrs`
        """
        for node_idx in self:
            node = self.nodes[node_idx]
            if attributes_match(node, attrs):
                yield node_idx

    def __getattr__(self, name):
        # TODO: DRY
        if name.startswith('get_') and name.endswith('s'):
            type_ = name[len('get_'):-len('s')]
            return partial(self.get_interaction, type_)
        elif name.startswith('add_'):
            type_ = name[len('add_'):]
            return partial(self.add_interaction, type_)
        elif name.startswith('remove_'):
            type_ = name[len('remove_'):]
            return partial(self.remove_interaction, type_)
        else:
            raise AttributeError('Unknown attribute "{}".'.format(name))

    def merge_molecule(self, molecule):
        """
        Add the atoms and the interactions of a molecule at the end of this
        one.

        Atom and residue index of the new atoms are offset to follow the last
        atom of this molecule.

        Parameters
        ----------
        molecule: Molecule
            The molecule to merge at the end.

        Returns
        -------
        dict
            A dict mapping the node indices of the added `molecule` to their
            new indices in this molecule.
        """
        if self.force_field != molecule.force_field:
            raise ValueError(
                'Cannot merge molecules with different force fields.'
            )
        if self.nrexcl is None and not self:
            self.nrexcl = molecule.nrexcl
        if self.nrexcl != molecule.nrexcl:
            raise ValueError(
                'Cannot merge molecules with different nrexcl. '
                'This molecule has nrexcl={}, while the other has nrexcl={}.'
                .format(self.nrexcl, molecule.nrexcl)
            )
        if self.nodes():
            # We assume that the last id is always the largest.
            last_node_idx = max(self)
            offset = last_node_idx
            residue_offset = self.nodes[last_node_idx].get('resid', 1)
            offset_charge_group = self.nodes[last_node_idx].get('charge_group', 1)
        else:
            offset = 0
            residue_offset = 0
            offset_charge_group = 0
        correspondence = {}
        for idx, node in enumerate(molecule.nodes(), start=offset + 1):
            correspondence[node] = idx
            new_atom = copy.copy(molecule.nodes[node])
            new_atom['resid'] = (new_atom.get('resid', 1) + residue_offset)
            new_atom['charge_group'] = (new_atom.get('charge_group', 1)
                                        + offset_charge_group)
            self.add_node(idx, **new_atom)
        for name, interactions in molecule.interactions.items():
            for interaction in interactions:
                atoms = tuple(correspondence[atom] for atom in interaction.atoms)
                #print(atoms, interaction.meta)
                self.add_interaction(name, atoms, interaction.parameters, interaction.meta)
        for node1, node2 in molecule.edges:
            if correspondence[node1] != correspondence[node2]:
                self.add_edge(correspondence[node1], correspondence[node2])
        return correspondence

    def share_moltype_with(self, other):
        """
        Checks whether `other` has the same shape as this molecule.

        Parameters
        ----------
        other: Molecule

        Returns
        -------
        bool
            True iff other has the same shape as this molecule.
        """
        # Almost identical to __eq__, except that some node attributes don't
        # contribute, such as position and chain.
        # Note that isomorphic molecules get separate moltypes now, since the
        # order of the nodes is coupled to the order of the atoms in the output
        # PDB
        ignore_attrs = ('position', 'chain', 'graph', 'mapping_weights')
        return (
            self.nrexcl == other.nrexcl and
            self._force_field == other._force_field and
            self.same_nodes(other, ignore_attr=ignore_attrs) and
            self.same_edges(other) and
            self.same_interactions(other)
        )

    # TODO: Allow comparison of interactions between isomorphic molecules.
    def same_interactions(self, other):
        """
        Returns `True` if the interactions are the same.

        To be equal, two interactions must share the same node key reference,
        the same interaction parameters, and the same meta attributes. Empty
        interaction categories are ignored.

        Parameters
        ----------
        other: Molecule

        Returns
        -------
        bool
        """
        keys_self = set(
            interaction_type
            for interaction_type, interactions
            in self.interactions.items()
            if interactions
        )
        keys_other = set(
            interaction_type
            for interaction_type, interactions
            in other.interactions.items()
            if interactions
        )
        # We first make sure that the two molecules share the same relevant
        # interaction categories. A relevant category is one that actually has
        # interactions. If the molecules share the relevant categories, then we
        # can loop over the categories of one or the other, without issues with
        # mismatches.
        if keys_self != keys_other:
            return False
        return all(
            self.interactions[interaction_type] == other.interactions[interaction_type]
            for interaction_type in keys_self
        )

    # TODO: Allow default values for attributes.
    # In most cases, we assume that an unspecified attribute is equivalent to
    # the attribute being None; we should allow for this in the comparison. We
    # should also allow an attribute to have an arbitrary default value. For
    # instance, the assumed default for the `PTM_atom` attribute is False.
    def same_nodes(self, other, ignore_attr=()):
        """
        Returns `True` if the nodes are the same and in the same order.

        The equality criteria used for the attribute values are those of
        :func:`vermouth.utils.are_different`.

        Parameters
        ----------
        other: Molecule
        ignore_attr: collections.abc.Container
            Attribute keys that will not be considered in the comparison.

        Returns
        -------
        bool
        """
        # We first check that the node keys match between the two molecules.
        # The order matters here, and we count on the fact that Molecules are
        # OrderedDicts.
        if list(self.nodes.keys()) != list(other.nodes.keys()):
            return False
        zipped_nodes = zip(self.nodes.values(), other.nodes.values())
        for self_node, other_node in zipped_nodes:
            # For each pair of nodes, we compare the attribute dictionaries.
            # The order does not matter here.
            self_keys = set(key for key in self_node if key not in ignore_attr)
            other_keys = set(key for key in other_node if key not in ignore_attr)
            if self_keys != other_keys:
                return False
            # We can loop over the keys because we tested above that they were
            # matching between the two dicts.
            for key in self_keys:
                self_value = self_node[key]
                other_value = other_node[key]
                if utils.are_different(self_value, other_value):
                    return False
        return True

    def same_edges(self, other):
        """
        Compare the edges between this molecule and an other.

        Edges are unordered and undirected, but they can have attributes.

        Parameters
        ----------
        other: networkx.Graph
            The other molecule to compare the edges with.

        Returns
        -------
        bool
        """
        # The items of `graph.edges(data=True)` are formatted as
        # (from, to, {dict of attributes}).
        edges_self = {
            frozenset(edge[:2]): edge[2]
            for edge in self.edges(data=True)
        }
        edges_other = {
            frozenset(edge[:2]): edge[2]
            for edge in other.edges(data=True)
        }
        # We first start with the cheapest test: if the edges do not match
        # regardless of attributes, we can save the more costly tests.
        if set(edges_self.keys()) != set(edges_other.keys()):
            return False

        # We know that the keys in `edges_self` and `edges_other` are the same.
        return all(
            not utils.are_different(edges_self[edge], edges_other[edge])
            for edge in edges_self
        )

    def iter_residues(self):
        """
        Returns a generator over the nodes of this molecules residues.

        Returns
        -------
        collections.abc.Generator
        """
        residue_graph = graph_utils.make_residue_graph(self)
        return (tuple(residue_graph.nodes[res]['graph'].nodes) for res in sorted(residue_graph.nodes))

    def edges_between(self, n_bunch1, n_bunch2, data=False):
        """
        Returns all edges in this molecule between nodes in `n_bunch1` and
        `n_bunch2`.

        Parameters
        ----------
        n_bunch1: :class:`~collections.abc.Iterable`
            The first bunch of node indices.
        n_bunch2: :class:`~collections.abc.Iterable`
            The second bunch of node indices.

        Returns
        -------
        :class:`list`
            A list of tuples of edges in this molecule. The first element of
            the tuple will be in `n_bunch1`, the second element in `n_bunch2`.
        """
        set_1 = set(n_bunch1)
        set_2 = set(n_bunch2)
        for node1 in set_1:
            cross = set_2 & set(self[node1])
            for node2 in cross:
                if not data:
                    yield (node1, node2)
                else:
                    yield (node1, node2, self.edges[node1, node2])

    def _remove_interactions_with_node(self, node):
        """
        We iterate through the different interactions we have and
        remove the interactions where the atoms to be deleted are present.
        Further we also delete the entire interaction_type if it is
        empty after all the necessary interactions have been deleted.
        """
        for name, interactions in self.interactions.items():
            # We *must* copy interactions (list call), otherwise you change
            # interactions while iterating over it, causing it to miss
            # consecutive interactions that should be removed.
            for interaction in list(interactions):
                if node in interaction.atoms:
                    self.interactions[name].remove(interaction)

        for interaction_type in list(self.interactions):
            if not self.interactions[interaction_type]:
                self.interactions.pop(interaction_type)

    def remove_node(self, node):
        """
        Overriding the remove_node method of networkx
        as we have to delete the interaction from the interactions list
        separately which is not a part of the graph and hence does not
        get deleted.
        """
        super().remove_node(node)
        self._remove_interactions_with_node(node)

    def remove_nodes_from(self, nodes):
        """
        Overriding the remove_nodes_from method of networkx
        as we have to delete the interaction from the
        interactions list separately which is not a part of
        the graph and hence does not get deleted.
        """
        super().remove_nodes_from(nodes)
        for node in nodes:
            self._remove_interactions_with_node(node)

    def make_edges_from_interaction_type(self, type_):
        """
        Create edges from the interactions of a given type.

        The interactions must be described so that two consecutive atoms in an
        interaction should be linked by an edge. This is the case for bonds,
        angles, proper dihedral angles, and cmap torsions. It is not always
        true for improper torsions.

        Cmap are described as two consecutive proper dihedral angles. The
        atoms for the interaction are the 4 atoms of the first dihedral angle
        followed by the next atom forming the second dihedral angle with the
        3 previous ones. Each pair of consecutive atoms generate an edge.

        .. warning::

            If there is no interaction of the required type, it will be
            silently ignored.

        Parameters
        ----------
        type_: str
            The name of the interaction type the edges should be built from.
        """
        for interaction in self.interactions.get(type_, []):
            if interaction.meta.get('edge', True):
                atoms = interaction.atoms
                self.add_edges_from(zip(atoms[:-1], atoms[1:]))

    def make_edges_from_interactions(self):
        """
        Create edges from the interactions we know how to convert to edges.

        The known interactions are bonds, angles, proper dihedral angles,
        cmap torsions and constraints.
        """
        known_types = ('bonds', 'angles', 'dihedrals', 'cmap', 'constraints')
        for type_ in known_types:
            self.make_edges_from_interaction_type(type_)

class Block(Molecule):
    """
    Residue topology template

    Two blocks are equal if the underlying molecules are equal, and if the
    block names are equal.

    Parameters
    ----------
    incoming_graph_data:
        Data to initialize graph. If None (default) an empty graph is created.
    attr:
        Attributes to add to graph as key=value pairs.

    Attributes
    ----------
    name: str or None
        The name of the residue. Set to `None` if undefined.
    interactions: dict
        All the known interactions. Each item of the dictionary is a type of
        interaction, with the key being the name of the kind of interaction
        using Gromacs itp/rtp conventions ('bonds', 'angles', ...) and the
        value being a list of all the interactions of that type in the residue.
        An interaction is a dict with a key 'atoms' under which is stored the
        list of the atoms involved (referred by their name), a key 'parameters'
        under which is stored an arbitrary list of non-atom parameters as
        written in a RTP file, and arbitrary keys to store custom metadata. A
        given interaction can have a comment under the key 'comment'.
    """
    # As the particles are stored as nodes, we want the nodes to stay
    # ordered.
    node_dict_factory = OrderedDict

    def __init__(self, incoming_graph_data=None, **attr):
        super(Block, self).__init__(incoming_graph_data, **attr)
        # Arbitrary attributes can be set during the initialization. We need
        # to set the default of some key attributes, but without overwritting
        # what has been passed in the 'attr' argument.
        defaults = {
            'name': None,
            'interactions': {},
        }
        self._set_defaults(defaults)
        self._apply_to_all_interactions = defaultdict(dict)

    def _set_defaults(self, defaults):
        for attribute, default_value in defaults.items():
            if not hasattr(self, attribute):
                setattr(self, attribute, default_value)

    def __eq__(self, other):
        return self.name == other.name and super().__eq__(other)

    def __repr__(self):
        name = self.name
        if name is None:
            name = 'Unnamed'
        return '<{} "{}" at 0x{:x}>'.format(self.__class__.__name__,
                                            name, id(self))

    def add_atom(self, atom):
        """
        Add an atom. `atom` must contain an 'atomname'. This value will be this
        atom's index.

        Parameters
        ----------
        atom: collections.abc.Mapping
            The attributes of the atom to add. Must contain 'atomname'

        Raises
        ------
        ValueError
            If `atom` does not contain 'atomname'
        """
        try:
            name = atom['atomname']
        except KeyError:
            raise ValueError('Atom has no atomname: "{}".'.format(atom))
        self.add_node(name, **atom)

    @property
    def atoms(self):
        """"
        The atoms in the residue. Each atom is a dict with *a minima* a key
        'name' for the name of the atom, and a key 'atype' for the atom type.
        An atom can also have a key 'charge', 'charge_group', 'comment', or any
        arbitrary key.

        Returns
        -------
        collections.abc.Iterator[dict]
        """
        for node in self.nodes():
            node_attr = self.nodes[node]
            # In pre-blocks, some nodes correspond to particles in neighboring
            # residues. These node do not carry particle information and should
            # not appear as particles.
            if node_attr:
                yield node_attr

    def guess_angles(self):
        """
        Generates all possible triplets of node indices that correspond to
        angles.

        Yields
        ------
        tuple[collections.abc.Hashable, collections.abc.Hashable, collections.abc.Hashable]
            All possible angles.
        """
        for a in self.nodes():
            for b in self.neighbors(a):
                for c in self.neighbors(b):
                    if c == a:
                        continue
                    yield (a, b, c)

    def guess_dihedrals(self, angles=None):
        """
        Generates all possible quadruplets of node indices that correspond to
        torsion angles.

        Parameters
        ----------
        angles: collections.abc.Iterable
            All possible angles from which to start looking for torsion angles.
            Generated from :meth:`guess_angles` if not provided.

        Yields
        ------
        tuple[collections.abc.Hashable, collections.abc.Hashable, collections.abc.Hashable, collections.abc.Hashable]
            All possible torsion angles.
        """
        angles = angles if angles is not None else self.guess_angles()
        for a, b, c in angles:
            for d in self.neighbors(c):
                if d not in (a, b):
                    yield (a, b, c, d)

    def has_dihedral_around(self, center):
        """
        Returns True if the block has a dihedral centered around the given bond.

        Parameters
        ----------
        center: tuple
            The name of the two central atoms of the dihedral angle. The
            method is sensitive to the order.

        Returns
        -------
        bool
        """
        all_centers = [tuple(dih['atoms'][1:-1])
                       for dih in self.interactions.get('dihedrals', [])]
        return tuple(center) in all_centers

    def has_improper_around(self, center):
        """
        Returns True if the block has an improper centered around the given bond.

        Parameters
        ----------
        center: tuple
            The name of the two central atoms of the improper torsion. The
            method is sensitive to the order.

        Returns
        -------
        bool
        """
        all_centers = [tuple(dih.atoms[1:-1])
                       for dih in self.interactions.get('impropers', [])]
        return tuple(center) in all_centers

    def to_molecule(self, atom_offset=0, offset_resid=0, offset_charge_group=0,
                    force_field=None, default_attributes=None):
        """
        Converts this block to a :class:`Molecule`.

        Parameters
        ----------
        atom_offset: int
            The number at which to start numbering the node indices.
        offset_resid: int
            The offset for the `resid` attributes.
        offset_charge_group: int
            The offset for the `charge_group` attributes.
        force_field: None or vermouth.forcefield.ForceField
        default_attributes: collections.abc.Mapping[str]
            Attributes to set to for nodes that are missing them.

        Returns
        -------
        Molecule
            This block as a molecule.
        """
        if force_field is None:
            force_field = self.force_field
        if default_attributes is None:
            default_attributes = {'resname': self.name}
        name_to_idx = {}
        mol = Molecule(force_field=force_field)
        for idx, node in enumerate(self.nodes, start=atom_offset):
            name_to_idx[node] = idx
            atom = self.nodes[node]
            new_atom = default_attributes.copy()
            new_atom.update(atom)
            new_atom['resid'] = (new_atom.get('resid', 1) + offset_resid)
            new_atom['charge_group'] = (new_atom.get('charge_group', 1) +
                                        offset_charge_group)
            mol.add_node(idx, **new_atom)
        for name, interactions in self.interactions.items():
            for interaction in interactions:
                atoms = tuple(
                    name_to_idx[atom] for atom in interaction.atoms
                )
                mol.add_interaction(
                    name, atoms,
                    interaction.parameters,
                    meta=interaction.meta
                )
        for edge in self.edges:
            mol.add_edge(*(name_to_idx[node] for node in edge))

        try:
            mol.nrexcl = self.nrexcl
        except AttributeError:
            pass

        return mol


class Link(Block):
    """
    Template link between two residues.

    Two links are equal if:

    * the underlying molecules are equal
    * the names are equal
    * the negative edges ("non-edges") are equal regardless of order
    * the interactions to remove are the same and in the same order
    * the meta variables are equal
    * the pattern definitions are equal and in the same order
    * the features are equals regardless of order

    A link does not match if any of the non-edges match the target; their
    order therefore is not important. Same goes for features that just need to
    be present or not. The order does matter however for interactions to remove
    as removing the interactions in a different order may lead to a different
    set of remaining interactions.

    Parameters
    ----------
    incoming_graph_data:
        Data to initialize graph. If `None` (default) an empty graph is created.
    attr:
        Attributes to add to graph as key=value pairs.
    """
    node_dict_factory = OrderedDict

    def __init__(self, incoming_graph_data=None, **attr):
        super().__init__(incoming_graph_data, **attr)
        # Arbitrary attributes can be set during the initialization. We need
        # to set the default of some key attributes, but without overwritting
        # what has been passed in the 'attr' argument.
        defaults = {
            'non_edges': [],
            'removed_interactions': {},
            'molecule_meta': {},
            'patterns': [],
            'features': set(),
        }
        self._set_defaults(defaults)
        self._apply_to_all_nodes = {}

    def __eq__(self, other):
        # TODO: There is no good reason to care about the order of patterns
        # except for the fact that order free comparison of non hashable things
        # is a pain.
        return (super().__eq__(other)
                and self.same_non_edges(other)
                and self.removed_interactions == other.removed_interactions
                and self.molecule_meta == other.molecule_meta
                and self.patterns == other.patterns
                and set(self.features) == set(other.features)
                )

    def same_non_edges(self, other):
        """
        Returns `True` if all the non-edges of an `other` link are equal to
        those of this link. Returns `False` otherwise.
        """
        # A non-edge is a list of two elements: the key to a node in the graph
        # that is used as anchor, and an attribute dict that must match none of
        # the atoms connected to the anchor.
        # For the link to match, none of the non-edges must match. Therefore,
        # their order do not matter. Though, because the attribute dicts are
        # not hashable, non-edges cannot be put in a set.

        # If the number of non-edges is not the same, we can save the hasle.
        if len(self.non_edges) != len(other.non_edges):
            return False
        # If there is no non-edges, which is likely the most common case, we
        # can also save some effort.
        if not self.non_edges:
            return True

        # We sort the non-edges to get rid of the order. However, we cannot
        # sort them on the attribute dict, and there may be more than one
        # non-edge involving a given anchor. So for each anchor, we need to
        # compare the dicts of that link with all the non already assigned
        # dicts of the other link that share the same anchor.
        sorted_self = sorted(self.non_edges, key=lambda x: x[0])
        sorted_other = sorted(other.non_edges, key=lambda x: x[0])
        zipped = zip(itertools.groupby(sorted_self, key=lambda x: x[0]),
                     itertools.groupby(sorted_other, key=lambda x: x[0]))
        for (key_self, attrs_self), (key_other, attrs_other) in zipped:
            if key_self != key_other:
                return False
            attrs_self = list(attrs_self)
            attrs_other = list(attrs_other)
            if len(attrs_self) != len(attrs_other):
                return False
            for attr_self in attrs_self:
                for idx, attr_other in enumerate(attrs_other):
                    if not utils.are_different(attr_self, attr_other):
                        del attrs_other[idx]
                        break
                else:  # no break
                    return False
        return True


def attributes_match(attributes, template_attributes, ignore_keys=()):
    """
    Compare a dict of attributes from a molecule with one from a link.

    Returns ``True`` if the attributes from the link match the ones from the
    molecule; returns ``False`` otherwise. The attributes from a link match
    with those of a molecule is all the individual attribute from the link
    match the corresponding ones in the molecule. In the simplest case, these
    attribute match if their values are equal. If the value of the link
    attribute is an instance of :class:`LinkPredicate`, then the attributes
    match if the ``match`` method of the predicate returns ``True``.

    Parameters
    ----------
    attributes: dict
        Attributes from the molecule.
    template_attributes: dict
        Attributes from the link.
    ignore_keys: list
        List of keys to ignore from 'template_attributes'.

    Returns
    -------
    bool
    """
    for attr, value in template_attributes.items():
        if attr in ignore_keys:
            continue
        if attributes.get(attr) != value:
            if isinstance(value, LinkPredicate) and value.match(attributes, attr):
                continue
            return False
    return True


def interaction_match(molecule, interaction, template_interaction):
    """
    Compare an interaction with a template interaction or interaction to delete.

    An instance of :class:`Interaction` matches a template instance of the same
    class or of :class:`DeleteInteraction` if, at the  minimum, it involves the
    same atoms in the same order. If the template defines parameters, then they
    have to match as well. In the case of of a :class:`DeleteInteraction`,
    atoms may have attributes as well, then they have to match with the
    attributes of the corresponding atoms in the molecule.

    Parameters
    ----------
    molecule: networkx.Graph
        The molecule that contains the interaction.
    interaction: Interaction
        The interaction in the molecule.
    template_interaction: Interaction or DeleteInteraction
        The template to match with the interaction.

    Returns
    -------
    bool

    See Also
    --------
    attributes_match
    """
    atoms_match = tuple(template_interaction.atoms) == tuple(interaction.atoms)
    parameters_match = (
        not template_interaction.parameters
        or tuple(template_interaction.parameters) == tuple(interaction.parameters)
    )
    if atoms_match and parameters_match:
        try:
            atom_attrs = template_interaction.atom_attrs
        except AttributeError:
            atom_attrs = [{}, ] * len(template_interaction.atoms)
        nodes = [molecule.nodes[atom] for atom in interaction.atoms]
        for atom, template_atom in zip(nodes, atom_attrs):
            if not attributes_match(atom, template_atom):
                return False
        return attributes_match(interaction.meta, template_interaction.meta)
    return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the ForceField class.
"""

import pathlib
import pytest
import vermouth.forcefield
import vermouth.molecule

# pylint: disable=redefined-outer-name

@pytest.fixture
def empty_force_field():
    """
    Build an empty force field named 'empty'.
    """
    return vermouth.forcefield.ForceField(name='empty')


@pytest.fixture
def force_field_with_features(empty_force_field):
    """
    Build a force field with links describing features.

    The force field is empty except for the links. The links are empty except
    for the feature lists.
    """
    features = (
        ['single_0'], ['single_1'], ['single_2'],
        ['single_0', 'multi_0', 'multi_1'],
        ['multi_0', 'multi_2'],
        [],
    )
    for feature_names in features:
        link = vermouth.molecule.Link()
        link.features = feature_names
        empty_force_field.links.append(link)
    return empty_force_field


@pytest.mark.parametrize('attribute, value', (
    ('name', 'empty'),
    ('blocks', {}),
    ('links', []),
    ('modifications', {}),
    ('renamed_residues', {}),
    ('variables', {}),
    ('reference_graphs', {}),
    ('features', set()),
))
def test_init_empty_force_field(empty_force_field, attribute, value):
    """
    Test that an empty force field has the expected attributes.
    """
    assert getattr(empty_force_field, attribute) == value


def test_features(force_field_with_features):
    """
    Test that link features are found by the force field.
    """
    expected = {'single_0', 'single_1', 'single_2',
                'multi_0', 'multi_1', 'multi_2'}
    assert force_field_with_features.features == expected


@pytest.mark.parametrize('name', (
    'single_0', 'single_1', 'single_2', 'multi_0', 'multi_1', 'multi_2'
))
def test_has_feature_true(force_field_with_features, name):
    """
    Test that :attr:`vermouth.forcefield.ForceField.has_feature` finds the
    expected features.
    """
    assert force_field_with_features.has_feature(name)


def test_has_feature_false(force_field_with_features):
    """
    Test that :attr:`vermouth.forcefield.ForceField.has_feature` does not find
    features that should not be there.
    """
    assert not force_field_with_features.has_feature('absent')


def test_create_ff_from_dir_name(tmpdir):
    """
    Creates a force field from a directory AND a name, assure that the correct
    name is used.

    The name provided with the 'name' argument should be used.
    """
    directory = tmpdir.mkdir('dirname')
    name = 'the_name'
    ff = vermouth.forcefield.ForceField(
        directory=str(directory),
        name=name,
    )
    assert ff.name == name


def test_create_ff_from_name():
    """
    Creates a force field by name (no directory) and make sure it has the
    expected name.
    """
    name = 'the_ff_name'
    ff = vermouth.forcefield.ForceField(name=name)
    assert ff.name == name


@pytest.mark.parametrize('path_type', (str, pathlib.Path))
def test_create_ff_from_dir(tmpdir, path_type):
    """
    Creates a force field from a directory, assure that the name is correct.
    """
    ff_name = 'name'
    # Depending on the version of python, the type of tmpdir may differ.
    directory = path_type(str(tmpdir.mkdir(ff_name)))
    ff = vermouth.forcefield.ForceField(directory=directory)
    assert ff.name == ff_name


def test_init_empty_force_field_error_neither():
    """
    Test that an error is raised when neither directory nor name are given to
    ForceField.
    """
    with pytest.raises(TypeError):
        vermouth.forcefield.ForceField()



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides several generic utility functions
"""

import collections.abc
import itertools
import numbers
import string

import numpy as np


# Do not define in the except so the function can be tested.
def _distance(point_1, point_2):
    """
    .. autofunction:: scipy.spatial.distance.euclidean
    """
    return np.sqrt(np.sum((point_1 - point_2)**2))


try:
    from scipy.spatial.distance import euclidean as distance  # pylint: disable=unused-import
except ImportError:
    distance = _distance


class _Filler:
    """
    Utility class for :func:`are_different`.

    An instance of this class is used as filler when comparing iterables that
    may not have the same length.
    """
    pass


def format_atom_string(node, atomid='', chain='', resname='', resid='', atomname=''):
    defaults = dict(atomid=atomid, chain=chain, resname=resname, resid=resid,
                    atomname=atomname)
    defaults.update(node)
    return '{atomid}{chain}-{resname}{resid}:{atomname}'.format(**defaults)


def maxes(iterable, key=lambda x: x):
    """
    Analogous to ``max``, but returns a list of all maxima.

    >>> all(key(elem) == max(iterable, key=key) for elem in iterable)
    True

    Parameters
    ----------
    iterable: collections.abc.Iterable
        The iterable for which to find all maxima.
    key: collections.abc.Callable
        This callable will be called on each element of ``iterable`` to evaluate
        it to a value. Return values must support ``>`` and ``==``.

    Returns
    -------
    list
        A list of all maximal values.

    """
    max_key = None
    out = []
    for item in iterable:
        key_val = key(item)
        if max_key is None or key_val > max_key:
            out = [item]
            max_key = key_val
        elif key_val == max_key:
            out.append(item)
    return out


def first_alpha(search_string):
    """
    Returns the first ASCII letter.

    Parameters
    ----------
    string: str
        The string in which to look for the first ASCII letter.

    Returns
    -------
    str

    Raises
    ------
    ValueError
        No ASCII letter was found in 'search_string'.
    """
    for elem in search_string:
        # str.isalpha catches all unicode charaters tagged as "letter"; it is a
        # very broad set of characters.
        if elem in string.ascii_letters:
            return elem
    raise ValueError('No alpha charecters in "{}".'.format(search_string))


def are_all_equal(iterable):
    """
    Returns ``True`` if and only if all elements in `iterable` are equal; and
    ``False`` otherwise.

    Parameters
    ----------
    iterable: collections.abc.Iterable
        The container whose elements will be checked.

    Returns
    -------
    bool
        ``True`` iff all elements in `iterable` compare equal, ``False``
        otherwise.
    """
    try:
        shape = iterable.shape
    except AttributeError:
        pass
    else:
        if len(shape) > 1:
            message = 'The function does not works on multidimension arrays.'
            raise NotImplementedError(message) from None

    iterator = iter(iterable)
    first = next(iterator, None)
    return all(np.all(item == first) for item in iterator)


def are_different(left, right):
    """
    Return True if two values are different from one another.

    Values are considered different if they do not share the same type. In case
    of numerical value, the comparison is done with :func:`numpy.isclose` to
    account for rounding. In the context of this test, `nan` compares equal to
    itself, which is not the default behavior.

    The order of mappings (dicts) is assumed to be irrelevant, so two
    dictionaries are not different if the only difference is the order of the
    keys.
    """
    if left.__class__ != right.__class__:
        return True

    # Because we know that `left` and `right` share the same type, we also know
    # that if `left` is `None`, then `right` is also `None`, so `left` and
    # `right` are NOT different. It is an easy and common case, so we treat it
    # early to avoid extra work.
    if left is None:
        return False

    if isinstance(left, numbers.Number):
        try:
            return not np.isclose(left, right, equal_nan=True)
        except TypeError:
            # Some things pretend to be numbers but cannot go through isclose.
            # It is the case of integers that overflow an int64 for instance.
            return left != right

    if isinstance(left, (str, bytes)):
        return left != right

    filler = _Filler()

    if isinstance(left, collections.abc.Mapping):
        left_key_set = set(left.keys())
        right_key_set = set(right.keys())
        if left_key_set != right_key_set:
            return True
        return any(are_different(left[key], right[key]) for key in left_key_set)

    if isinstance(left, collections.abc.Iterable):
        zipped = itertools.zip_longest(left, right, fillvalue=filler)
        return any(
            item_left is filler or
            item_right is filler or
            are_different(item_left, item_right)
            for item_left, item_right in zipped
        )

    return left != right



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a class to describe a system.
"""


class System:
    """
    A system of molecules.

    Attributes
    ----------
    molecules: list[:class:`~vermouth.molecule.Molecule`]
        The molecules in the system.
    """
    def __init__(self, force_field=None):
        self.molecules = []
        self._force_field = None
        self.force_field = force_field

    @property
    def force_field(self):
        """
        The forcefield used to describe the molecules in this system.
        """
        return self._force_field

    @force_field.setter
    def force_field(self, value):
        """
        Set the forcefield for all the molecules in the system.

        Parameters
        ----------
        value: :class:`~vermouth.forcefield.ForceField`
        """
        self._force_field = value
        for molecule in self.molecules:
            molecule._force_field = value  # pylint: disable=protected-access

    def add_molecule(self, molecule):
        """
        Add a molecule to the system.

        Parameters
        ----------
        molecule: :class:`~vermouth.molecule.Molecule`
        """
        self.molecules.append(molecule)

    @property
    def num_particles(self):
        """
        The total number of particles in all the molecules in this system.
        """
        return sum(len(mol) for mol in self.molecules)

    def copy(self):
        """
        Creates a copy of this system and it's molecules.

        Returns
        -------
        System
            A deep copy of this system.
        """
        new_system = self.__class__()
        new_system.molecules = [mol.copy() for mol in self.molecules]
        new_system.force_field = self.force_field
        return new_system



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy

import numpy as np
import pytest

from vermouth import ffinput
from vermouth.molecule import Choice, Link, Block
from vermouth import forcefield



CHOICE = Choice(['A', 'B'])


@pytest.mark.parametrize('key, ref_prefix, ref_base', (
    ('BB', '', 'BB'),
    ('+XX', '+', 'XX'),
    ('--UNK', '--', 'UNK'),
    ('>>>PLOP', '>>>', 'PLOP'),
    ('<<KEY', '<<', 'KEY'),
    ('++ONE+TWO', '++', 'ONE+TWO'),
    ('*****NODE', '*****', 'NODE'),
))
def test_split_node_key(key, ref_prefix, ref_base):
    """
    Test _split_node_key works as expected.
    """
    prefix, base = ffinput._split_node_key(key)
    assert prefix == ref_prefix
    assert base == ref_base


@pytest.mark.parametrize('key', (
    '',  # empty key
    '++',  # no base
    '++-BB',  # mixed prefix
))
def test_split_node_key_error(key):
    """
    Test _split_node_key raises an error when expected.
    """
    with pytest.raises(IOError):
        prefix, base = ffinput._split_node_key(key)
        print('prefix: "{}"; base: "{}"'.format(prefix, base))


@pytest.mark.parametrize('attributes, ref_prefix, ref_order', (
    ({}, '', None),
    ({'order': None}, '', None),
    ({'arbitrary': 'something'}, '', None),
    ({'atomname': 'BB'}, '', None),
    ({'order': 4}, '++++', 4),
    ({'order': -2}, '--', -2),
    ({'order': 0}, '', 0),
    ({'order': np.int16(4)}, '++++', 4),
    ({'order': '>'}, '>', '>'),
    ({'order': '>>'}, '>>', '>>'),
    ({'order': '<'}, '<', '<'),
    ({'order': '<<<'}, '<<<', '<<<'),
    ({'order': '*'}, '*', '*'),
    ({'order': '****'}, '****', '****'),
))
def test_get_order_and_prefix_from_attributes(attributes, ref_prefix, ref_order):
    """
    Test _get_order_and_prefix_from_attributes works as expected.
    """
    (result_prefix,
     result_order) = ffinput._get_order_and_prefix_from_attributes(attributes)
    assert result_prefix == ref_prefix
    assert result_order == ref_order


@pytest.mark.parametrize('attributes', (
    {'order': ''},
    {'order': 4.5},
    {'order': True},
    {'order': False},
    {'order': np.True_},
    {'order': '++'},
    {'order': '-'},
    {'order': '><'},
    {'order': 'invalid'},
))
def test_get_order_and_prefix_from_attributes_error(attributes):
    """
    Test _get_order_and_prefix_from_attributes fails when expected.
    """
    with pytest.raises(IOError):
        result = ffinput._get_order_and_prefix_from_attributes(attributes)
        print(result)


@pytest.mark.parametrize('prefix, ref_prefix, ref_order', (
    ('', None, 0),
    ('+', '+', 1),
    ('++', '++', 2),
    ('-', '-', -1),
    ('---', '---', -3),
    ('>', '>', '>'),
    ('>>>>', '>>>>', '>>>>'),
    ('<', '<', '<'),
    ('<<<', '<<<', '<<<'),
    ('*', '*', '*'),
    ('**', '**', '**'),
))
def test_get_order_and_prefix_from_prefix(prefix, ref_prefix, ref_order):
    """
    Test _get_order_and_prefix_from_prefix works as expected.
    """
    result_prefix, result_order = ffinput._get_order_and_prefix_from_prefix(prefix)
    assert result_prefix == ref_prefix
    assert result_order == ref_order


@pytest.mark.parametrize('key, attributes, ref_prefixed, ref_attributes', (
    # No attribute given
    ('BB', {}, 'BB', {'atomname': 'BB', 'order': 0}),
    ('+BB', {}, '+BB', {'atomname': 'BB', 'order': 1}),
    ('--BB', {}, '--BB', {'atomname': 'BB', 'order': -2}),
    ('>>XX', {}, '>>XX', {'atomname': 'XX', 'order': '>>'}),
    ('<<<YYY', {}, '<<<YYY', {'atomname': 'YYY', 'order': '<<<'}),
    ('****Z', {}, '****Z', {'atomname': 'Z', 'order': '****'}),
    # Order given by the attribute
    ('BB', {'order': 3}, '+++BB', {'atomname': 'BB', 'order': 3}),
    ('XX', {'order': -4}, '----XX', {'atomname': 'XX', 'order': -4}),
    ('YYY', {'order': '>'}, '>YYY', {'atomname': 'YYY', 'order': '>'}),
    ('ZZZZ', {'order': '<<<'}, '<<<ZZZZ', {'atomname': 'ZZZZ', 'order': '<<<'}),
    ('A', {'order': '**'}, '**A', {'atomname': 'A', 'order': '**'}),
    # Order given both with prefix and attribute
    ('BB', {'order': 0}, 'BB', {'atomname': 'BB', 'order': 0}),
    ('++BB', {'order': 2}, '++BB', {'atomname': 'BB', 'order': 2}),
    ('--BB', {'order': -2}, '--BB', {'atomname': 'BB', 'order': -2}),
    ('>>BB', {'order': '>>'}, '>>BB', {'atomname': 'BB', 'order': '>>'}),
    ('<<BB', {'order': '<<'}, '<<BB', {'atomname': 'BB', 'order': '<<'}),
    ('**BB', {'order': '**'}, '**BB', {'atomname': 'BB', 'order': '**'}),
    # atomname given in attribute
    ('BB', {'atomname': 'BB'}, 'BB', {'atomname': 'BB', 'order': 0}),
    ('>>XX', {'atomname': 'BB'}, '>>XX', {'atomname': 'BB', 'order': '>>'}),
    ('YYY', {'atomname': 'BB', 'order': '**'}, '**YYY', {'atomname': 'BB', 'order': '**'}),
    # CHOICE is `Choice(['A', 'B'])`, it is an instance of `molecule.LinkPredicate`.
    ('XX', {'atomname': CHOICE}, 'XX', {'atomname': CHOICE, 'order': 0}),
))
def test_treat_atom_prefix(key, attributes,
                           ref_prefixed, ref_attributes):
    """
    Test _treat_atom_prefix works as expected.
    """
    (result_prefixed,
     result_attributes) = ffinput._treat_atom_prefix(key, attributes)
    assert result_prefixed == ref_prefixed
    assert result_attributes == ref_attributes
    assert result_attributes is not attributes  # We do a shallow copy


def test_treat_atom_prefix_error():
    """
    Test _treat_atom_prefix fails when order is inconsistent.
    """
    with pytest.raises(IOError):
        ffinput._treat_atom_prefix('+A', {'order': -1})


@pytest.mark.parametrize('line, tokens', (
    ('2  3  1  0.2  1000', ['2', '3', '1', '0.2', '1000']),
    ('PO4  GL1  1 0.2  1000', ['PO4', 'GL1', '1', '0.2', '1000']),
    ('PO4  GL1  --  1 0.2  1000', ['PO4', 'GL1', '--', '1', '0.2', '1000']),
    (
        "BB {'resname': 'ALA', 'secstruc': 'H'} BB "
        "{'resname': 'LYS', 'secstruc': 'H', 'order': +1} 1 0.2 1000",
        ['BB', "{'resname': 'ALA', 'secstruc': 'H'}", 'BB',
         "{'resname': 'LYS', 'secstruc': 'H', 'order': +1}",
         '1', '0.2', '1000']
    ),
    (
        "BB {'resname': 'ALA', 'secstruc': 'H'} "
        "+BB {'resname': 'LYS', 'secstruc': 'H'} -- 1 0.2 1000",
        ['BB', "{'resname': 'ALA', 'secstruc': 'H'}",
         '+BB', "{'resname': 'LYS', 'secstruc': 'H'}",
         '--', '1', '0.2', '1000']
    ),
    ('ATOM1{attributes}ATOM2', ['ATOM1', '{attributes}', 'ATOM2']),
    ('ATOM1 {attributes} ATOM2', ['ATOM1', '{attributes}', 'ATOM2']),
    # Note: the brackets here are NOT for str.format!
    ('{} {{}} {{}{{}}}', ['{}', '{{}}', '{{}{{}}}']),  # not str.format
    ('{}{{}}{{}{{}}}', ['{}', '{{}}', '{{}{{}}}']),  # not str.format
    ('', []),
    ('    ', []),
    ('\t', []),
))
def test_tokenize(line, tokens):
    """
    Test that _tokenize works as expected.
    """
    found = ffinput._tokenize(line)
    assert found == tokens


@pytest.mark.parametrize('token', (
    '{{{}}',  # missing closing bracket
    '{{}}}',  # missing openning bracket
))
def test_tokenize_bracket_error(token):
    """
    Test that _tokenize recognizes missing brackets.
    """
    with pytest.raises(IOError):
        # One closing bracket is missing.
        ffinput._tokenize(token)


@pytest.mark.parametrize('line, macros, expected', (
    (  # base case
        'content $macro other content',
        {'macro': 'plop'},
        'content plop other content',
    ),
    (  # non-separated macros
        'content $macro$other end',
        {'macro': 'plop', 'other': 'toto'},
        'content ploptoto end',
    ),
    ('hop $macro{toto}', {'macro': 'plop'}, 'hop plop{toto}'),
    ('hop $macro\n', {'macro': 'plop'}, 'hop plop\n'),
    ('hop $macro\t', {'macro': 'plop'}, 'hop plop\t'),
    ('hop $macro', {'macro': 'plop'}, 'hop plop'),
    (  # nothing to replace
        'hop hop hop',
        {'macro': 'plop'},
        'hop hop hop',
    ),
    (  # empty string
        '',
        {'macro': 'plop'},
        '',
    ),
    (  # multiple times the same
        'content $macro $macro content',
        {'macro': 'plop'},
        'content plop plop content',
    ),
    ('$macro', {'macro': 'plop'}, 'plop'),  # only the macro
    ('start $macro', {'macro': 'plop'}, 'start plop'),  # end with macro
    ('$macro end', {'macro': 'plop'}, 'plop end'),  # start with macro
    ('A {$key : "$value"} B',  # there *must* be a space between the
                               # macro and the colon, otherwise the key is
                               # 'key:' with the colon as a colon is not a
                               # token separator.
     {'key': '"a_key"', 'value': 'a_value'},
     'A {"a_key" : "a_value"} B'),  # play with json dicts
))
def test_substitute_macros(line, macros, expected):
    """
    Test _substitute_macros works as expected.
    """
    found = ffinput._substitute_macros(line, macros)
    assert found == expected


def test_substitute_macros_missing():
    """
    Test _substitute_macros fails when a macro is missing.
    """
    with pytest.raises(KeyError):
        ffinput._substitute_macros('hop $missing end', {'macro': 'plop'})


@pytest.mark.parametrize('tokens, atoms, natoms, expected', (
    (['A', 'B', 'C'], [], 2, True),
    (['A', 'B', 'C'], [], 0, False),
    (['A', 'B', 'C'], ['X', 'Y'], 3, True),
    (['A', 'B', 'C'], ['X', 'Y'], 2, False),
    (['A', 'B', '--', 'C'], [], None, True),
    (['--', 'A', 'B', 'C'], [], None, False),
    ([], [], None, False),
))
def test_some_atoms_left(tokens, atoms, natoms, expected):
    """
    Test _some_atoms_left works as expected.
    """
    tokens = collections.deque(tokens)
    found = ffinput._some_atoms_left(tokens, atoms, natoms)
    assert found == expected


@pytest.mark.parametrize('token, expected', (
    ('{}', {}),
    ('{"atomname": "PO4"}', {'atomname': 'PO4'}),
    ('{"a": "abc", "b": "def"}', {'a': 'abc', 'b': 'def'}),
    ('{"a": "123"}', {'a': '123'}),
    ('{"a": 123}', {'a': 123}),
    ('{"a": null}', {'a': None}),
    ('{"a": true}', {'a': True}),
    ('{"a": false}', {'a': False}),
    ('{"a": "A|B|C"}', {'a': Choice(['A', 'B', 'C'])}),
    ('{"a": {"b": "123"}}', {'a': {'b': '123'}}),
))
def test_parse_atom_attributes(token, expected):
    """
    Test _parse_atom_attributes works as expected.
    """
    found = ffinput._parse_atom_attributes(token)
    assert found == expected


@pytest.mark.parametrize('token', (
    '[1, 2]', '1', 'true', 'false', 'null',  # not dict-like
    "{'a': 123}", '{"a", "b"}',
))
def test_parse_atom_attributes_error(token):
    """
    Test _parse_atom_attributes fails as expected.
    """
    with pytest.raises(ValueError):
        ffinput._parse_atom_attributes(token)


@pytest.mark.parametrize('tokens, natoms, expected', (
    (['A', 'B', 'C'], None, [['A', {}], ['B', {}], ['C', {}]]),
    (['A', 'B'], 2, [['A', {}], ['B', {}]]),
    (['A', '{"a": "abc"}', 'B'], 2, [['A', {'a': 'abc'}], ['B', {}]]),
    (
        ['A', '{"a": "abc"}', 'B', '{"b": 0}'],
        2,
        [['A', {'a': 'abc'}], ['B', {'b': 0}]],
    ),
    (
        ['A', '{"a": "abc"}', 'B', '{"b": 0}', 'C'],
        2,
        [['A', {'a': 'abc'}], ['B', {'b': 0}]],
    ),
    (['1', '2', '1', '0.31', '7500'], 2, [['1', {}], ['2', {}]]),
    (
        ['BB', '{"replace": {"atype": "Qd", "charge": 1}}'],
        1,
        [['BB', {'replace': {'atype': 'Qd', 'charge': 1}}]],
    ),
    (
        ['BB', '++BB', '1', '0.640', '2500', '{"g": "Short", "edge": false}'],
        2,
        [['BB', {}], ['++BB', {}]],
    ),
    (
        ['-BB', 'BB', '+BB', '{"c": "C"}'],
        None,
        [['-BB', {}], ['BB', {}], ['+BB', {'c': 'C'}]],
    ),
    (
        ['-BB', '{"a": 1}', 'BB', '{"b": 2}', '+BB', '{"c": 3}'],
        None,
        [['-BB', {'a': 1}], ['BB', {'b': 2}], ['+BB', {'c': 3}]],
    ),
    (
        ['4', '6', '5', '--', '2'],
        None,
        [['4', {}], ['6', {}], ['5', {}]],
    ),
    (
        ['--'],
        None,
        [],
    ),
))
def test_get_atoms(tokens, natoms, expected):
    """
    Test _get_atoms works as expected.
    """
    tokens = collections.deque(tokens)
    remaining = collections.deque(tokens)
    found = ffinput._get_atoms(tokens, natoms)
    assert found == expected
    # Make sure '--' is consumed if needed
    if tokens:
        assert not tokens[0] == '--'


@pytest.mark.parametrize('token_list, remaining', (
    (['XX', 'YY', '--', 'AA', '--', 'BB'], ['AA', '--', 'BB']),
    (['A', 'B', '--', '--', '1', '2', '--', '3'], ['--', '1', '2', '--', '3']),
    (['A', 'B', '--'], []),
    (['--', 'A', 'B'], ['A', 'B']),
))
def test_separator(token_list, remaining):
    """
    _get_atoms only remove the appropriate separator.
    """
    tokens = collections.deque(token_list)
    ffinput._get_atoms(tokens, None)
    assert list(tokens) == remaining


@pytest.mark.parametrize('tokens, natoms', (
    (['{"a": "abc"}', 'A'], None),  # atom attributes without an atom
))
def test_get_atoms_errors(tokens, natoms):
    """
    Test _get_atoms fails when expected.
    """
    tokens = collections.deque(tokens)
    with pytest.raises(IOError):
        ffinput._get_atoms(tokens, natoms)


@pytest.mark.parametrize('atoms, apply_to_all, existing, expected', (
    ((), {}, {}, {}),
    (
        (('A', {}), ),
        {},
        {},
        {'A': {'atomname': 'A', 'order': 0}}
    ),
    (
        (('A', {}), ('B', {}), ),
        {},
        {},
        {'A': {'atomname': 'A', 'order': 0}, 'B': {'atomname': 'B', 'order':0}},
    ),
    (
        (('+A', {}), ),
        {},
        {},
        {'+A': {'atomname': 'A', 'order': 1}},
    ),
    (
        (('A', {}), ('B', {}), ),
        {'attr': 0},
        {},
        {
            'A': {'atomname': 'A', 'order': 0, 'attr': 0},
            'B': {'atomname': 'B', 'order': 0, 'attr': 0},
        }
    ),
    (
        (('A', {'exist': 'hello'}), ('B', {}), ),
        {},
        {'A': {'new': 'world'}},
        {
            'A': {'atomname': 'A', 'order': 0, 'exist': 'hello', 'new': 'world'},
            'B': {'atomname': 'B', 'order': 0},
        },
    ),
    (
        (('A', {'exist': 'hello'}), ('B', {}), ),
        {'attr': 'plop'},
        {'A': {'new': 'world'}},
        {
            'A': {
                'atomname': 'A', 'order': 0,
                'exist': 'hello', 'new': 'world',
                'attr': 'plop',
            },
            'B': {'atomname': 'B', 'order': 0, 'attr': 'plop'},
        },
    ),
    (
        (('A', {'exist': 'hello'}), ('B', {}), ),
        {'attr': 'plop', 'attr2': 'other plop'},
        {'A': {'new': 'world', 'new2': 'globe'}},
        {
            'A': {
                'atomname': 'A', 'order': 0,
                'exist': 'hello', 'new': 'world', 'new2': 'globe',
                'attr': 'plop', 'attr2': 'other plop',
            },
            'B': {
                'atomname': 'B', 'order': 0,
                'attr': 'plop', 'attr2': 'other plop',
            },
        },
    ),
))
def test_treat_link_interaction_atoms(atoms, apply_to_all, existing, expected):
    """
    Test that _treat_link_interaction_atoms works as expected.
    """
    context = Link()
    if apply_to_all:
        context._apply_to_all_nodes = apply_to_all
    context.add_nodes_from(existing.items())
    ffinput._treat_link_interaction_atoms(atoms, context, 'section')
    found = dict(context.nodes.items())
    assert found == expected


def test_treat_link_interaction_atoms_conflicts():
    """
    Test that _treat_link_interaction_atoms fails when there is a conflict
    between the atoms to add ans the existing atoms.
    """
    context = Link()
    context.add_nodes_from({
        'A': {'exist': 'before'},
    }.items())
    atoms = (('A', {'exist': 'after'}), )
    with pytest.raises(IOError):
        ffinput._treat_link_interaction_atoms(atoms, context, 'section')


@pytest.mark.parametrize('token, expected', (
    ('', False),
    ('(something)', False),
    ('before(inside)after', False),
    ('something()', True),
    ('something(inside)', True),
    ('partial(end', False),
    ('partialend)', False),
    ('(something)(or_other)', False),
    ('(something)or_other', False),
))
def test_is_param_effector(token, expected):
    """
    Test that _is_param_effector recognizes param effector tokens.
    """
    found = ffinput._is_param_effector(token)
    assert found == expected


@pytest.mark.parametrize('tokens, existing', (
    (['A'], set()),
    (['A', 'B'], set()),
    (['A', 'B'], {'X', 'Y'}),
    ([], set()),
    ([], {'X', 'Y'}),
))
def test_parse_features(tokens, existing):
    """
    Test that _parse_features works as expected.
    """
    context = Link()
    if existing:
        context.features = existing
    expected = existing | set(tokens)
    ffinput._parse_features(tokens, context, 'link')
    assert context.features == expected


def test_parse_features_wrong_type():
    """
    Test that _parse_features complains for context different than 'link'.
    """
    context = Link()
    with pytest.raises(IOError):
        ffinput._parse_features(['A', 'B'], context, 'not-link')


@pytest.mark.parametrize('tokens, expected', (
    (['A'], [['A', {}]]),
    (['A', 'B'], [['A', {}], ['B', {}]]),
    (['+A', '{"attr": 0}', 'B'], [['+A', {'attr': 0}], ['B', {}]]),
))
def test_parse_patterns(tokens, expected):
    """
    Test that _parse_patterns works as expected.
    """
    existing = [[['X', {}], ['Y', {}]]]
    full_expected = copy.copy(existing)
    full_expected.append(expected)
    tokens = collections.deque(tokens)
    context = Link()
    context.patterns = existing
    ffinput._parse_patterns(tokens, context, 'link')
    assert context.patterns == full_expected


def test_parse_patterns_wrong_type():
    """
    Test that _parse_patterns complains for context different than 'link'.
    """
    context = Link()
    with pytest.raises(IOError):
        ffinput._parse_patterns(['A', 'B'], context, 'not-link')


@pytest.mark.parametrize('tokens, expected', (
    (['name', '"value"'], {'name': 'value'}),
    (['name', '"a|b|c"'], {'name': 'a|b|c'}),
    (['name', 'value'], {'name': 'value'}),
    (['other-name', '0'], {'other-name': 0}),
    (
        ['name', '{"key": 123, "key2": "value"}'],
        {'name': {'key': 123, 'key2': 'value'}},
    ),
))
def test_parse_variables(tokens, expected):
    """
    Test that _parse_variables works as expected.
    """
    existing = {'existing': 'I was there'}
    full_expected = copy.copy(existing)
    full_expected.update(expected)
    force_field = forcefield.ForceField('dummy')
    force_field.variables = existing

    ffinput._parse_variables(tokens, force_field, 'section')

    assert force_field.variables == full_expected


@pytest.mark.parametrize('tokens', (
    [],
    ['single'],
    ['one', 'two', 'extra'],
))
def test_parse_variables_wrong_length(tokens):
    """
    Test that _parse_variables fails when the number of tokens is wrong.
    """
    force_field = forcefield.ForceField('dummy')
    with pytest.raises(IOError):
        ffinput._parse_variables(tokens, force_field, 'section')


@pytest.mark.parametrize('tokens', (
    ['key', 'value'],
    ['key1', 'new-value'],
    ['key', '"quoted-value"'],
))
def test_parse_macro(tokens):
    """
    Test that _parse_macro works as expected.
    """
    macros = {'key1': 'value1', 'key2': 'value2'}
    expected = copy.copy(macros)
    expected[tokens[0]] = tokens[1]
    tokens = collections.deque(tokens)
    ffinput._parse_macro(tokens, macros)
    assert macros == expected


@pytest.mark.parametrize('tokens', (
    [],
    ['one'],
    ['one', 'two', 'three'],
))
def test_parse_macro_wrong_length(tokens):
    """
    Test that _parse_macro fails when the number of tokens is wrong.
    """
    tokens = collections.deque(tokens)
    with pytest.raises(IOError):
        ffinput._parse_macro(tokens, {})


@pytest.mark.parametrize('tokens, expected', (
    (['#meta', '{"a": "123", "b": 123}'], {'a': '123', 'b': 123}),
    (['ignored', '{"a": "123", "b": 123}'], {'a': '123', 'b': 123}),
))
@pytest.mark.parametrize('context_class, context_type', (
    (Link, 'link'),
    (Block, 'block'),
    (Link, 'modification'),
))
def test_parse_meta(tokens, expected, context_class, context_type):
    """
    Test that _parse_meta works as expected.
    """
    existing = {'key': 'value'}

    full_expected = copy.copy(existing)
    full_expected.update(expected)

    context = context_class()
    context._apply_to_all_interactions['section'] = existing

    ffinput._parse_meta(tokens, context, context_type, 'section')
    assert context._apply_to_all_interactions['section'] == full_expected


@pytest.mark.parametrize('tokens', (
    [],
    ['one'],
    ['one', 'two', 'three'],
))
def test_parse_meta_wrong_length(tokens):
    """
    Test that _parse_meta fails when the number of tokens is wrong.
    """
    with pytest.raises(IOError):
        ffinput._parse_meta(tokens, None, 'type', 'section')



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created on Tue Aug 22 11:48:46 2017

@author: Peter Kroon
"""
#import matplotlib.pyplot as plt
#plt.close('all')
from vermouth import *

import os.path as osp

import numpy as np

#PATH = '../molecules/cycliclipopeptide_2.pdb'
#PATH = '../molecules/cyclicpeptide_2.pdb'
PATH = '../molecules/glkfk.pdb'
#PATH = '../molecules/6-macro-8_cartwheel.gro'
#PATH = '../molecules/6-macro-16.gro'
#PATH = '../molecules/6-macro-16-rtc-eq-nodisre.pdb'
#PATH = '../molecules/3-macro-1.gro'
#
#write_pdb(CG_graph, "6-macro-16-rtc-eq-nodisre-CG.pdb", conect=True)

system = System()

ext = osp.splitext(PATH)[-1]

if ext.casefold() == '.pdb':
    PDBInput().run_system(system, PATH)
elif ext.casefold() == '.gro':
    GROInput().run_system(system, PATH)
else:
    raise RuntimeError
MakeBonds().run_system(system)
RepairGraph().run_system(system)
CanonizePTMs().run_system(system)
DoMapping().run_system(system)
ApplyBlocks().run_system(system)
DoLinks().run_system(system)

print(system)


for mol in system.molecules:
    for idx in list(mol):
        if 'position' not in mol.nodes[idx]:
            node = mol.nodes[idx]
            print(node['resname'], node['resid'], node['atomname'])
#            mol.nodes[idx]['position'] = np.array([np.nan, np.nan, np.nan])
            mol.remove_node(idx)
            
    draw(mol, node_size=30, node_color=tuple(np.random.rand(3)), with_label=True)

show()



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unittests for the PDB reader.
"""

import numpy as np
import pytest

from vermouth.pdb.pdb import PDBParser
from vermouth.tests.datafiles import PDB_MULTIMODEL


def single_model(with_conect):
    """
    Provide a PDB representation of a single molecule.
    """
    pdb = '''
HETATM    1  C01 UNK     1       3.939  -2.095  -4.491  0.00  0.00           C   
HETATM    2  C02 UNK     1       4.825  -2.409  -5.534  0.00  0.00           C   
HETATM    3  C03 UNK     1       5.466  -3.658  -5.560  0.00  0.00           C   
HETATM    4  C04 UNK     1       5.207  -4.601  -4.552  0.00  0.00           C   
HETATM    5  C05 UNK     1       4.320  -4.288  -3.508  0.00  0.00           C   
HETATM    6  C06 UNK     1       3.693  -3.031  -3.473  0.00  0.00           C   
HETATM    7  C07 UNK     1       2.736  -2.681  -2.319  0.00  0.00           C   
HETATM    8  C08 UNK     1       2.386  -3.959  -1.535  0.00  0.00           C   
HETATM    9  C09 UNK     1       1.429  -3.608  -0.380  0.00  0.00           C   
HETATM   10  C10 UNK     1       1.079  -4.886   0.404  0.00  0.00           C   
HETATM   11  H01 UNK     1       3.217  -1.965  -1.652  0.00  0.00           H   
HETATM   12  H02 UNK     1       1.905  -4.675  -2.201  0.00  0.00           H   
HETATM   13  H03 UNK     1       1.824  -2.243  -2.724  0.00  0.00           H   
HETATM   14  H04 UNK     1       1.910  -2.892   0.287  0.00  0.00           H   
HETATM   15  H05 UNK     1       3.298  -4.397  -1.129  0.00  0.00           H   
HETATM   16  H06 UNK     1       0.598  -5.602  -0.262  0.00  0.00           H   
HETATM   17  H07 UNK     1       3.435  -1.113  -4.469  0.00  0.00           H   
HETATM   18  H08 UNK     1       5.018  -1.673  -6.334  0.00  0.00           H   
HETATM   19  H09 UNK     1       6.172  -3.900  -6.373  0.00  0.00           H   
HETATM   20  H10 UNK     1       5.699  -5.588  -4.579  0.00  0.00           H   
HETATM   21  H11 UNK     1       4.116  -5.028  -2.715  0.00  0.00           H   
HETATM   22  H12 UNK     1       0.517  -3.170  -0.786  0.00  0.00           H   
HETATM   23  H13 UNK     1       0.401  -4.638   1.221  0.00  0.00           H   
HETATM   24  H14 UNK     1       1.991  -5.324   0.810  0.00  0.00           H   
'''
    if with_conect:
        pdb += '''
CONECT    1    2                                                                
CONECT    1    2                                                                
CONECT    1    6   17                                                           
CONECT    2    1                                                                
CONECT    2    1                                                                
CONECT    2    3   18                                                           
CONECT    3    2                                                                
CONECT    3    4                                                                
CONECT    3    4                                                                
CONECT    3   19                                                                
CONECT    4    3                                                                
CONECT    4    3                                                                
CONECT    4    5   20                                                           
CONECT    5    4                                                                
CONECT    5    6                                                                
CONECT    5    6                                                                
CONECT    5   21                                                                
CONECT    6    1                                                                
CONECT    6    5                                                                
CONECT    6    5                                                                
CONECT    6    7                                                                
CONECT    7    6    8   11   13                                                 
CONECT    8    7    9   12   15                                                 
CONECT    9    8   10   14   22                                                 
CONECT   10    9   16   23   24                                                 
CONECT   11    7                                                                
CONECT   12    8                                                                
CONECT   13    7                                                                
CONECT   14    9                                                                
CONECT   15    8                                                                
CONECT   16   10                                                                
CONECT   17    1                                                                
CONECT   18    2                                                                
CONECT   19    3                                                                
CONECT   20    4                                                                
CONECT   21    5                                                                
CONECT   22    9                                                                
CONECT   23   10                                                                
CONECT   24   10                                                                
'''
    pdb += 'END'
    return pdb


def multi_mol(with_conect):
    pdb = '''
ATOM      1  C01 UNK     1       3.939  -2.095  -4.491  0.00  0.00           C2+   
ATOM      2  C02 UNK     1       4.825  -2.409  -5.534  0.00  0.00              
TER       3      UNK     1
ATOM      4  C03 UNK     1       5.466  -3.658  -5.560  0.00  0.00           C   
ATOM      5  C04 UNK     1       5.207  -4.601  -4.552  0.00  0.00           C   
ATOM      6  C05 UNK     1       5.207  -4.601  -4.552  0.00  0.00           C   
TER       7      UNK     1
'''
    if with_conect:
        pdb += '''
CONECT    1    2                                                                
CONECT    5    4    6    
'''
    return pdb


def merged_mol(with_conect):
    pdb = '''
ATOM      1  C01 UNK     1       3.939  -2.095  -4.491  0.00  0.00           C   
ATOM      2  C02 UNK     1       4.825  -2.409  -5.534  0.00  0.00           C   
TER       3      UNK     1
ATOM      4  C03 UNK     1       5.466  -3.658  -5.560  0.00  0.00           C   
ATOM      5  C04 UNK     1       5.207  -4.601  -4.552  0.00  0.00           C   
ATOM      6  C05 UNK     1       5.207  -4.601  -4.552  0.00  0.00           C   
TER       7      UNK     1
'''
    if with_conect:
        pdb += '''
CONECT    1    2                                                                
CONECT    5    4    6    2
'''
    return pdb


@pytest.mark.parametrize('pdbstr, ignh, nnodesnedges', (
    [single_model(True), False, ((24, 24),)],
    [single_model(False), False, ((24, 0),)],
    [single_model(True), True, ((10, 10),)],
    [single_model(False), True, ((10, 0),)],
    [multi_mol(True), None, ((2, 1), (3, 2))],
    [multi_mol(False), None, ((2, 0), (3, 0))],
    [merged_mol(True), None, ((5, 4),)],
    [merged_mol(False), None, ((2, 0), (3, 0))],
))
def test_single_model(pdbstr, ignh, nnodesnedges):
    parser = PDBParser(ignh=ignh)
    mols = list(parser.parse(pdbstr.splitlines()))
    assert len(mols) == len(nnodesnedges)
    for mol, nnodes_edges in zip(mols, nnodesnedges):
        nnodes, nedges = nnodes_edges
        assert len(mol.nodes) == nnodes
        assert len(mol.edges) == nedges


@pytest.mark.parametrize('ignh', [True, False])
@pytest.mark.parametrize('modelidx', range(1, 16))
def test_integrative(ignh, modelidx):
    parser = PDBParser(ignh=ignh, modelidx=modelidx)
    with open(str(PDB_MULTIMODEL)) as pdb_file:
        mols = list(parser.parse(pdb_file))
    assert len(mols) == 3  # 3 chains
    for mol in mols:
        if ignh:
            assert len(mol.nodes) == 441
            assert len(mol.edges) == 0  # No CONECT records
        else:
            assert len(mol.nodes) == 904
            assert len(mol.edges) == 0


def test_altloc(caplog):
    pdbstr = """
ATOM   1300  CA AHIS A 194     -13.902 -22.133  70.272  0.56 53.66
ATOM   1301  CA BHIS A 194     -13.910 -22.208  70.255  0.44 53.81
"""
    parser = PDBParser()
    mols = list(parser.parse(pdbstr.splitlines()))
    assert len(mols) == 1
    mol = mols[0]
    assert len(mol.nodes) == 1
    assert any(rec.levelname == 'WARNING' for rec in caplog.records)


def test_atom_attributes():
    """
    Test that atom attributes are parsed and set correctly
    """
    pdb = multi_mol(False)
    parser = PDBParser()
    mols = list(parser.parse(pdb.splitlines()))
    assert len(mols) == 2
    nodes = [
        {
            0: {'atomname': 'C01', 'atomid': 1, 'resid': 1,
                'resname': 'UNK', 'chain':'', 'altloc': '',
                'insertion_code': '', 'occupancy': 0.0, 'temp_factor': 0.0,
                'position': np.array([0.3939,  -0.2095,  -0.4491]),
                'charge': 2.0, 'element': 'C'},
            1: {'atomname': 'C02', 'atomid': 2, 'resid': 1,
                'resname': 'UNK', 'chain':'', 'altloc': '',
                'insertion_code': '', 'occupancy': 0.0, 'temp_factor': 0.0,
                'position': np.array([0.4825,  -0.2409,  -0.5534]),
                'charge': 0, 'element': 'C'}
        },
        {
            0: {'atomname': 'C03', 'atomid': 4, 'resid': 1,
                'resname': 'UNK', 'chain': '', 'altloc': '',
                'insertion_code': '', 'occupancy': 0.0, 'temp_factor': 0.0,
                'position': np.array([0.5466,  -0.3658,  -0.5560]),
                'charge': 0, 'element': 'C'},
            1: {'atomname': 'C04', 'atomid': 5, 'resid': 1,
                'resname': 'UNK', 'chain': '', 'altloc': '',
                'insertion_code': '', 'occupancy': 0.0, 'temp_factor': 0.0,
                'position': np.array([0.5207,  -0.4601,  -0.4552]),
                'charge': 0, 'element': 'C'},
            2: {'atomname': 'C05', 'atomid': 6, 'resid': 1,
                'resname': 'UNK', 'chain': '', 'altloc': '',
                'insertion_code': '', 'occupancy': 0.0, 'temp_factor': 0.0,
                'position': np.array([0.5207, -0.4601, -0.4552]),
                'charge': 0, 'element': 'C'}
        }
    ]
    for mol, n_attrs in zip(mols, nodes):
        for n_idx in mol.nodes:
            assert set(n_attrs[n_idx].keys()) == set(mol.nodes[n_idx].keys())
            for attr in mol.nodes[n_idx]:
                assert attr in n_attrs[n_idx]
                if attr == 'position':
                    assert np.allclose(n_attrs[n_idx][attr], mol.nodes[n_idx][attr])
                else:
                    assert n_attrs[n_idx][attr] == mol.nodes[n_idx][attr]



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the modifications made to the redistributed KDTree.
"""


import hypothesis
from hypothesis import strategies as st
import hypothesis.extra.numpy as hnp
import numpy as np
import pytest
from vermouth.redistributed.kdtree import KDTree as redisKDTree

try:
    from scipy.spatial import cKDTree as scipyKDTree
except ImportError:
    # scipy is not available
    HAS_SCIPY = False
else:
    HAS_SCIPY = True


def dict_close(left, right):
    """
    Compare 2 dicts whith floats as values.

    Returns `True` is all the keys are the same and the values are all close
    according to :func:`np.allclose`.

    The order of the keys is not kept.
    """
    if left.keys() != right.keys():
        return False
    keys = sorted(left)
    values_left = np.array([left[key] for key in keys])
    values_right = np.array([right[key] for key in keys])
    return np.allclose(values_left, values_right)


@pytest.mark.skipif(not HAS_SCIPY, reason="Scipy is not available.")
@hypothesis.given(
    coordinates=hnp.arrays(
        dtype=st.sampled_from((np.float32, np.float64)),
        shape=st.tuples(st.integers(1, 10), st.integers(1, 10)),
        elements=st.floats(allow_nan=False, allow_infinity=False, width=32),
    ),
    max_dist=st.floats(0, 5, width=32),
    p=st.integers(1, 5),
)
def test_sparse_distance_matrix(coordinates, max_dist, p):
    """
    Test that the `sparse_distance_matrix` method returns the same thing in the
    redistributed KDTree and on the actual scipy one.
    """
    original_tree = scipyKDTree(coordinates)
    redis_tree = redisKDTree(coordinates)

    original_output = original_tree.sparse_distance_matrix(original_tree, max_dist, p)
    redis_output = redis_tree.sparse_distance_matrix(redis_tree, max_dist, p)

    assert dict_close(redis_output, original_output)



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the selectors in :mod:`vermouth.selectors`.
"""

import functools

import pytest

import vermouth
from vermouth.pdb.pdb import read_pdb
from vermouth.tests.datafiles import (
    PDB_PROTEIN,
    PDB_NOT_PROTEIN,
    PDB_PARTIALLY_PROTEIN,
)


@pytest.mark.parametrize(
    'molecules, reference_answer',
    [(read_pdb(str(path)), answer) for path, answer in [
        (PDB_PROTEIN, True),
        (PDB_NOT_PROTEIN, False),
        (PDB_PARTIALLY_PROTEIN, False),
    ]]
)
def test_is_protein(molecules, reference_answer):
    """
    Make sure that proteins are correctly identified as such.
    """
    assert len(molecules) == 1
    molecule = molecules[0]
    assert vermouth.selectors.is_protein(molecule) == reference_answer


@pytest.mark.parametrize(
    'atom, reference_answer',
    (
        ({'position': [0, 0, 0]}, True),
        ({'position': None}, False),
        ({}, False),
    )
)
def test_selector_no_position(atom, reference_answer):
    """
    Test that :func:`vermouth.selectors.selector_has_position` works as expected.
    """
    assert vermouth.selectors.selector_has_position(atom) == reference_answer


@pytest.mark.parametrize('node', (
    {},
    {'resname': 'PLOP', 'atomname': 'A'},
))
def test_select_all(node):
    """
    Test that :func:`vermouth.selectors.select_all` indeed accept every nodes.
    """
    assert vermouth.selectors.select_all(node)


@pytest.mark.parametrize('node, expected', (
    ({}, False),  # empty node
    ({'atomname': 'not BB'}, False),  # wrong atomname
    ({'resname': 'anything'}, False),  # no atomname
    ({'atomname': 'BB'}, True),  # only correct atomname
    ({'atomname': 'BB', 'other': 'something'}, True),  # correct atomname + other attribute
))
def test_select_backbone(node, expected):
    """
    Test that :func:`vermouth.selectors.select_backbone` only select beads with
    'BB' as atom name.
    """
    assert vermouth.selectors.select_backbone(node) == expected


@pytest.mark.parametrize('node, attribute, values, expected', (
    ({}, 'name', ['something', 'other'], False),
    ({'name': 'nothing'}, 'name', ['something', 'other'], False),
    ({'name': 'something'}, 'name', ['something', 'other'], True),
    ({'name': 'nothing', 'plop': 'other'}, 'name', ['something', 'other'], False),
    ({'name': 'nothing'}, 'name', [], False),
))
def test_select_proto_select_attribute_in(node, attribute, values, expected):
    """
    Test that :func:`vermouth.selectors.proto_select_attribute_in` works as
    expected and can be used with :func:`functools.partial`.
    """
    # The use of functools.partial is not necessary. Though this is how
    # proto_select_attribute_in is used in the example from its docstring so
    # it should work.
    select_attribute_in = functools.partial(
        vermouth.selectors.proto_select_attribute_in,
        attribute=attribute, values=values,
    )
    assert select_attribute_in(node) == expected


@pytest.mark.parametrize('node, templates, ignore_keys, expected', (
    ({}, [], (), False),  # Everything empty
    ({}, [{'name': 'A'}, {'name': 'B'}], (), False),  # Empty node
    ({'name': 'A'}, [{'resname': 'A'}, {'resid': 1}], (), False),  # Not matching
    ({'name': 'A'}, [{'name': 'not A'}, {'name': 'A', 'resname': 'something'}], (), False),
    ({'name': 'A'}, [{'resid': 1, 'resname': 'something'}, {'name': 'A'}], (), True),
    ({'name': 'A', 'resname': 'plop'}, [{'name': 'A'}], ('resname', 'other'), True),
))
def test_proto_multi_templates(node, templates, ignore_keys, expected):
    """
    Test that :func:`vermouth.selectors.proto_multi_templates` works as
    expected and can be used with :func:`functools.partial`.
    """
    multi_templates = functools.partial(
        vermouth.selectors.proto_multi_templates,
        templates=templates,
        ignore_keys=ignore_keys,
    )
    assert multi_templates(node) == expected


def test_filter_minimal():
    """
    Test that :func:`vermouth.selectors.filter_minimal` works as expected.
    """
    # Build a molecule that has all even atoms with no positions.
    molecules = read_pdb(str(PDB_PROTEIN))
    assert len(molecules) == 1
    molecule = molecules[0]

    for atom in list(molecule.nodes.values())[::2]:
        atom['position'] = None
    # This means that we want to keep the odd atoms
    to_keep = list(molecule.nodes)[1::2]

    filtered = vermouth.selectors.filter_minimal(
        molecule,
        selector=vermouth.selectors.selector_has_position
    )

    # Do we keep the right atoms?
    assert list(filtered) == to_keep



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides a processor that sorts atoms within molecules.
"""

import copy
from functools import partial
from .processor import Processor


class SortMoleculeAtoms(Processor):
    """
    Sort the atoms within a molecule by chain, resid, and resname.

    This is usefull, for instance, when atoms have been added (*e.g.* missing
    atoms identified by :class:`vermouth.processors.repair_graph.RepairGraph`).
    The atom keys are left identical, only the order of the nodes is changed.
    """
    def run_molecule(self, molecule):
        node_order = sorted(molecule, key=partial(_keyfunc, molecule))
        sorted_nodes = [
            (node_key, molecule.nodes[node_key])
            for node_key in node_order
        ]
        edges = tuple(molecule.edges(data=True))
        interactions = copy.copy(molecule.interactions)
        molecule.remove_nodes_from(node_order)
        molecule.add_nodes_from(sorted_nodes)
        molecule.add_edges_from(edges)
        molecule.interactions = interactions
        return molecule


def _keyfunc(graph, node_idx):
    """
    Reduce a molecule node to a tuple of chain, resid, and resname.
    """
    # TODO add something like idx_in_residue
    return (
        graph.nodes[node_idx]['chain'],
        graph.nodes[node_idx]['resid'],
        graph.nodes[node_idx]['resname'],
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Processor adding edges between molecules.
"""

from .processor import Processor
from .. import edge_tuning
from ..molecule import Choice

# TODO: this should be defined in the force fields
DNA_DONORS = (
    {'resname': Choice(['DA', 'DA3', 'DA5']), 'atomname': Choice(['C2', 'N6'])},
    {'resname': Choice(['DG', 'DG3', 'DG5']), 'atomname': Choice(['N1', 'N2'])},
    {'resname': Choice(['DC', 'DC3', 'DC5']), 'atomname': 'N4'},
    {'resname': Choice(['DT', 'DT3', 'DT5']), 'atomname': 'N3'},
)
DNA_ACCEPTORS = (
    {'resname': Choice(['DA', 'DA3', 'DA5']), 'atomname': 'N1'},
    {'resname': Choice(['DG', 'DG3', 'DG5']), 'atomname': 'O6'},
    {'resname': Choice(['DC', 'DC3', 'DC5']), 'atomname': Choice(['N3', 'O2'])},
    {'resname': Choice(['DT', 'DT3', 'DT5']), 'atomname': Choice(['O2', 'O4'])},
)
DNA_HB_DIST = 0.30


class AddMoleculeEdgesAtDistance(Processor):
    """
    Processor that adds edges within and between molecules.

    The processor adds edges between atoms, within or between molecules, when
    the atoms are part of the selections provided for each end of the edges,
    and the atoms are closer than a given threshold.

    Parameters
    ----------
    threshold: float
        Distance threshold in nanometers under which to create an edge.
    templates_from: list[dict]
        List of node templates to select the atoms at one end of the edges.
    templates_to: list[dict]
        List of node template to select the atoms at the other end of the edges.
    attribute: str
        Name of the attribute under which are stores the coordinates.

    See Also
    --------
    vermouth.molecule.attributes_match
    """
    def __init__(self, threshold, templates_from, templates_to,
                 attribute='position', min_edges=0):
        self.threshold = threshold
        self.templates_from = templates_from
        self.templates_to = templates_to
        self.attribute = attribute
        self.min_edges = min_edges

    def run_system(self, system):
        """
        Run the processor on the system.
        """
        system.molecules = edge_tuning.add_edges_threshold(
            system.molecules,
            threshold=self.threshold,
            templates_a=self.templates_from,
            templates_b=self.templates_to,
            attribute=self.attribute,
            min_edges=self.min_edges,
        )
        return system


class MergeNucleicStrands(AddMoleculeEdgesAtDistance):
    """
    Add edges between complementary nucleic acid strands.

    By default, the edges are added in place of the hydrogen bonds between
    complementary bases.

    Parameters
    ----------
    threshold: float
        Distance threshold in nanometers under which to create an edge.
    templates_donnors: list[dict]
        List of templates describing hydrogen donnors.
    templates_acceptors: list[dict]
        List of templates describing hydrogen acceptors.
    attribute: str
        Name of the attribute under which are store the node coordinates.
    """
    def __init__(self, threshold=DNA_HB_DIST,
                 templates_donnors=DNA_DONORS,
                 templates_acceptors=DNA_ACCEPTORS,
                 attribute='position'):
        super().__init__(
            threshold=threshold,
            templates_from=templates_donnors,
            templates_to=templates_acceptors,
            attribute=attribute,
            min_edges=4
        )



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides functions for reading and writing PDB files.
"""

import numpy as np
import networkx as nx

from ..file_writer import open
from ..molecule import Molecule
from ..utils import first_alpha, distance, format_atom_string
from ..parser_utils import LineParser
from ..truncating_formatter import TruncFormatter
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


class PDBParser(LineParser):
    """
    Parser for PDB files

    Attributes
    ----------
    active_molecule: vermouth.molecule.Molecule
        The molecule/model currently being read.
    molecules: list[vermouth.molecule.Molecule]
        All complete molecules read so far.
    modelidx: int
        Which model to take.

    Parameters
    ----------
    exclude: collections.abc.Container[str]
        Container of residue names. Any atom that has a residue name that is in
        `exclude` will be skipped.
    ignh: bool
        Whether all hydrogen atoms should be skipped
    modelidx: int
        Which model to take.
    """

    def __init__(self, exclude=('SOL',), ignh=False, modelidx=1):
        self.active_molecule = Molecule()
        self.molecules = []
        self._conects = []
        self.exclude = exclude
        self.ignh = ignh
        self.modelidx = modelidx
        self._skipahead = False

    def dispatch(self, line):
        """
        Returns the appropriate method for parsing `line`. This is determined
        based on the first 6 characters of `line`.

        Parameters
        ----------
        line: str

        Returns
        -------
        collections.abc.Callable[str, int]
            The method to call with the line, and the line number.
        """
        record = line[:6].strip().lower()
        return getattr(self, record, self._unknown_line)

    def parse(self, file_handle):
        # Only PDBParser.finalize should produce a result, namely a list of
        # molecules. This means that mols is a list containing a single list of
        # molecules, which is a little silly.
        outcome = list(super().parse(file_handle))
        assert len(outcome) == 1
        yield from outcome[0]

    @staticmethod
    def _unknown_line(line, lineno):
        """
        Called when a line is unknown. Raises a KeyError with a helpful message.
        """
        raise KeyError("Line {} can not be parsed, since we don't recognize the"
                       " first 6 characters. The line is: {}"
                       "".format(lineno, line))

    @staticmethod
    def _skip(line, lineno=0):
        """
        Does nothing.
        """


    # TODO: Parse some of these, and either do something useful with it, or
    #       propagate it to e.g. the ITP
    # pylint: disable=bad-whitespace
    # TITLE SECTION
    header = _skip
    obslte = _skip
    title  = _skip
    splt   = _skip
    caveat = _skip
    compnd = _skip
    source = _skip
    keywds = _skip
    expdta = _skip
    nummdl = _skip
    mdltyp = _skip
    author = _skip
    revdat = _skip
    sprsde = _skip
    jrnl   = _skip
    remark = _skip

    # PRIMARY STRUCTURE SECTION
    dbref  = _skip
    dbref1 = _skip
    dbref2 = _skip
    seqadv = _skip
    seqres = _skip
    modres = _skip

    # HETEROGEN SECTION
    het    = _skip
    formul = _skip
    hetnam = _skip
    hetsyn = _skip

    # SECONDARY STRUCTURE SECTION
    helix  = _skip
    sheet  = _skip

    # CONNECTIVITY ANNOTATION SECTION
    ssbond = _skip
    link   = _skip
    cispep = _skip

    # MISCELLANEOUS FEATURES SECTION
    site   = _skip

    # CRYSTALLOGRAPHIC AND COORDINATE TRANSFORMATION SECTION
    cryst1 = _skip
    origx1 = _skip
    origx2 = _skip
    origx3 = _skip
    scale1 = _skip
    scale2 = _skip
    scale3 = _skip
    mtrix1 = _skip
    mtrix2 = _skip
    mtrix3 = _skip

    # COORDINATE SECTION
    # model  = _skip  # Used
    # atom   = _skip  # Used
    anisou = _skip
    # ter    = _skip  # Used
    # hetatm = _skip  # Used
    # endmdl = _skip  # Used

    # CONNECTIVITY SECTION
    # conect = _skip  # Used

    # BOOKKEEPING SECTION
    master = _skip
    # end   = _skip  # Used
    # pylint: enable=bad-whitespace

    def _atom(self, line, lineno=0):
        """
        Parse an ATOM or HETATM record.

        Parameters
        ----------
        line: str
            The line to parse. We do not check whether it starts with either
            "ATOM  " or "HETATM".
        lineno: int
            The line number (not used).
        """
        if self._skipahead:
            return

        fields = [
            ('', str, 6),
            ('atomid', int, 5),
            ('', str, 1),
            ('atomname', str, 4),
            ('altloc', str, 1),
            ('resname', str, 4),
            ('chain', str, 1),
            ('resid', int, 4),
            ('insertion_code', str, 1),
            ('', str, 3),
            ('x', float, 8),
            ('y', float, 8),
            ('z', float, 8),
            ('occupancy', float, 6),
            ('temp_factor', float, 6),
            ('', str, 10),
            ('element', str, 2),
            ('charge', str, 2),
        ]

        start = 0
        field_slices = []
        for name, type_, width in fields:
            if name:
                field_slices.append((name, type_, slice(start, start + width)))
            start += width

        properties = {}
        for name, type_, slice_ in field_slices:
            value = line[slice_].strip()
            if value:
                properties[name] = type_(value)
            else:
                properties[name] = type_()

        # Charge is special, since it's "2-" or "1+", rather than -2 etc. And
        # let's turn it into a number
        charge = properties['charge']
        if charge:
            try:
                charge = float(charge)
            except ValueError:
                charge = float(charge[::-1])
        else:
            charge = 0
        properties['charge'] = charge

        pos = (properties.pop('x'), properties.pop('y'), properties.pop('z'))
        # Coordinates are read in Angstrom, but we want them in nm
        properties['position'] = np.array(pos, dtype=float) / 10

        if not properties['element']:
            atomname = properties['atomname']
            properties['element'] = first_alpha(atomname)
        if properties['altloc'] not in ['', 'A']:
            # TODO: allow selecting alternative conformation for specific
            #       residues.
            LOGGER.warning("There is an alternative conformation for atom {}. "
                           "We use conformation A exclusively",
                           format_atom_string(properties),
                           type='pdb-alternate')
            return
        if (properties['resname'] in self.exclude or
                (self.ignh and properties['element'] == 'H')):
            return
        idx = max(self.active_molecule) + 1 if self.active_molecule else 0
        self.active_molecule.add_node(idx, **properties)

    atom = _atom
    hetatm = _atom

    def model(self, line, lineno=0):
        """
        Parse a MODEL record. If the model is not the same as :attr:`modelidx`,
        this model will not be parsed.

        Parameters
        ----------
        line: str
            The line to parse. Should start with "MODEL ", but this is not
            checked.
        lineno: int
            The line number (not used).
        """
        try:
            modelnr = int(line[10:14])
        except ValueError:
            return
        else:
            self._skipahead = modelnr != self.modelidx

    def conect(self, line, lineno=0):
        """
        Parse a CONECT record. The line is stored for later processing.

        Parameters
        ----------
        line: str
            The line to parse. Should start with CONECT, but this is not checked
        lineno: int
            The line number (not used).
        """
        # We can't add edges immediately, since the molecule might not be parsed
        # yet (does the PDB file format mandate anything on the order of
        # records?). Instead, just store the lines for later use.
        self._conects.append(line)

    def _finish_molecule(self, line="", lineno=0):
        """
        Finish parsing the molecule. :attr:`active_molecule` will be appended to
        :attr:`molecules`, and a new :attr:`active_molecule` will be made.
        """
        # We kind of *want* to yield self.active_molecule here, but we can't
        # since there's a very good chance it's CONECT records have not been
        # parsed yet, and the molecule won't have any edges.
        if self.active_molecule:
            self.molecules.append(self.active_molecule)
        self.active_molecule = Molecule()

    endmdl = _finish_molecule
    ter = _finish_molecule
    end = _finish_molecule

    def finalize(self, lineno=0):
        """
        Finish parsing the file. Process all CONECT records found, and returns
        a list of molecules.

        Parameters
        ----------
        lineno: int
            The line number (not used).

        Returns
        -------
        list[vermouth.molecule.Molecule]
            All molecules parsed from this file.
        """
        # TODO: cross reference number of molecules with CMPND records
        self._finish_molecule()
        self.do_conect()
        return self.molecules

    def do_conect(self):
        """
        Apply connections to molecule based on CONECT records read from PDB file
        """
        id2idxs = [{mol.nodes[idx]['atomid']: idx for idx in mol}
                   for mol in self.molecules]
        for line in self._conects:
            start = 6
            width = 5
            atids = []
            for num in range(start, len(line.rstrip()), width):
                atom = int(line[num:num + width])
                atids.append(atom)
            self._do_single_conect(atids, id2idxs)

    def _do_single_conect(self, conect_record, id2idxs):
        """
        Process a single CONECT line. Adds edges to the molecules in
        :attr:`molecules`.

        Parameters
        ----------
        conect_record: list[int]
        id2idxs: list[dict[int, int]]
            A list of dicts mapping atomids to node keys for every molecule in
            :attr:`molecules`
        """
        atomid0 = conect_record[0]
        # Find the appropriate molecule:
        mol = None
        for mol, id2idx in zip(self.molecules, id2idxs):
            if atomid0 in id2idx:
                atomidx0 = id2idx[atomid0]
                break
        else:  # no break
            # No molecule found with an atom containing atomid atid0.
            # This could be a hydrogen if ignh, or an excluded residue.
            return
        for atomid in conect_record[1:]:
            # Find the second molecule...
            mol2 = None
            for mol2, id2idx in zip(self.molecules, id2idxs):
                if atomid in id2idx:
                    atomidx = id2idx[atomid]
                    break
            else:
                # Two options: a skipped atom
                continue
            if mol is not mol2:
                assert mol is not None and mol2 is not None
                LOGGER.info('Merging two molecules/chains because there is a '
                            'CONECT record between atoms {} and {}',
                            format_atom_string(mol.nodes[atomidx0]),
                            format_atom_string(mol2.nodes[atomidx]))
                # Conect record between two molecules! It's probably a *good*
                # idea to cross reference this with e.g. SSBOND and LINK records
                molidx = self.molecules.index(mol)
                molidx2 = self.molecules.index(mol2)
                del id2idxs[max(molidx, molidx2)]
                del id2idxs[min(molidx, molidx2)]
                self.molecules.remove(mol)
                self.molecules.remove(mol2)
                mol = nx.disjoint_union(mol, mol2)
                mol2 = mol
                self.molecules.append(mol)
                id2idxs.append({mol.nodes[idx]['atomid']: idx for idx in mol})

            dist = distance(mol.nodes[atomidx0]['position'],
                            mol2.nodes[atomidx]['position'])
            mol.add_edge(atomidx0, atomidx, distance=dist)


def read_pdb(file_name, exclude=('SOL',), ignh=False, modelidx=1):
    """
    Parse a PDB file to create a molecule.

    Parameters
    ----------
    filename: str
        The file to read.
    exclude: collections.abc.Container[str]
        Atoms that have one of these residue names will not be included.
    ignh: bool
        Whether hydrogen atoms should be ignored.
    model: int
        If the PDB file contains multiple models, which one to select.

    Returns
    -------
    vermouth.molecule.Molecule
        The parsed molecules. Will only contain edges if the PDB file has
        CONECT records. Either way, might be disconnected.
    """
    parser = PDBParser(exclude, ignh, modelidx)
    with open(str(file_name)) as file_handle:
        mols = list(parser.parse(file_handle))
    LOGGER.info('Read {} molecules from PDB file {}', len(mols), file_name)
    return mols


def get_not_none(node, attr, default):
    """
    Returns ``node[attr]``. If it doesn't exists or is ``None``, return
    `default`.

    Parameters
    ----------
    node: collections.abc.Mapping
    attr: collections.abc.Hashable
    default
        The value to return if ``node[attr]`` is either ``None``, or does not
        exist.

    Returns
    -------
    object
        The value of ``node[attr]`` if it exists and is not ``None``, else
        `default`.
    """
    value = node.get(attr)
    if value is None:
        value = default
    return value


def write_pdb_string(system, conect=True, omit_charges=True, nan_missing_pos=False):
    """
    Describes `system` as a PDB formatted string. Will create CONECT records
    from the edges in the molecules in `system` iff `conect` is True.

    Parameters
    ----------
    system: vermouth.system.System
        The system to write.
    conect: bool
        Whether to write CONECT records for the edges.
    omit_charges: bool
        Whether charges should be omitted. This is usually a good idea since
        the PDB format can only deal with integer charges.
    nan_missing_pos: bool
        Wether the writing should fail if an atom does not have a position.
        When set to `True`, atoms without coordinates will be written
        with 'nan' as coordinates; this will cause the output file to be
        *invalid* for most uses.

    Returns
    -------
    str
        The system as PDB formatted string.
    """
    out = []

    formatter = TruncFormatter()
#    format_string = 'ATOM  {: >5.5d} {:4.4s}{:1.1s}{:3.3s} {:1.1s}{:4.4d}{:1.1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          {:2.2s}{:2.2s}'
    format_string = 'ATOM  {: >5dt} {:4st}{:1st}{:3st} {:1st}{:>4dt}{:1st}   {:8.3ft}{:8.3ft}{:8.3ft}{:6.2ft}{:6.2ft}          {:2st}{:2st}'

    nodeidx2atomid = {}
    atomid = 1
    for mol_idx, molecule in enumerate(system.molecules):
        for node_idx in molecule:
            # Node indices do not have to be unique across molecules. So store
            # them as (mol_idx, node_idx)
            nodeidx2atomid[(mol_idx, node_idx)] = atomid
            node = molecule.nodes[node_idx]
            atomname = get_not_none(node, 'atomname', '')
            altloc = get_not_none(node, 'altloc', '')
            resname = get_not_none(node, 'resname', '')
            chain = get_not_none(node, 'chain', '')
            resid = get_not_none(node, 'resid', 1)
            insertion_code = get_not_none(node, 'insertioncode', '')
            try:
                # converting from nm to A
                x, y, z = node['position'] * 10  # pylint: disable=invalid-name
            except KeyError:
                if nan_missing_pos:
                    x = y = z = float('nan')  # pylint: disable=invalid-name
                else:
                    raise
            occupancy = get_not_none(node, 'occupancy', 1)
            temp_factor = get_not_none(node, 'temp_factor', 0)
            element = get_not_none(node, 'element', '')
            charge = get_not_none(node, 'charge', 0)
            if charge and not omit_charges:
                charge = '{:+2d}'.format(int(charge))[::-1]
            else:
                charge = ''
            line = formatter.format(format_string, atomid, atomname, altloc,
                                    resname, chain, resid, insertion_code, x,
                                    y, z, occupancy, temp_factor, element,
                                    charge)
            atomid += 1
            out.append(line)
        terline = formatter.format('TER   {: >5dt}      {:3st} {:1st}{: >4dt}{:1st}',
                                   atomid, resname, chain, resid, insertion_code)
        atomid += 1
        out.append(terline)
    if conect:
        number_fmt = '{:>4dt}'
        format_string = 'CONECT '
        for mol_idx, molecule in enumerate(system.molecules):
            for node_idx in molecule:
                todo = sorted(nodeidx2atomid[(mol_idx, n_idx)]
                              for n_idx in molecule[node_idx] if n_idx > node_idx)
                while todo:
                    current, todo = todo[:4], todo[4:]
                    fmt = ['CONECT'] + [number_fmt]*(len(current) + 1)
                    fmt = ' '.join(fmt)
                    line = formatter.format(fmt, nodeidx2atomid[(mol_idx, node_idx)], *current)
                    out.append(line)
    out.append('END   ')
    return '\n'.join(out)


def write_pdb(system, path, conect=True, omit_charges=True, nan_missing_pos=False):
    """
    Writes `system` to `path` as a PDB formatted string.

    Parameters
    ----------
    system: vermouth.system.System
        The system to write.
    path: str
        The file to write to.
    conect: bool
        Whether to write CONECT records for the edges.
    omit_charges: bool
        Whether charges should be omitted. This is usually a good idea since
        the PDB format can only deal with integer charges.
    nan_missing_pos: bool
        Whether the writing should fail if an atom does not have a position.
        When set to `True`, atoms without coordinates will be written
        with 'nan' as coordinates; this will cause the output file to be
        *invalid* for most uses.
        for most use.

    See Also
    --------
    :func:write_pdb_string
    """
    with open(path, 'w') as out:
        out.write(write_pdb_string(system, conect, omit_charges, nan_missing_pos))



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains integrative tests for the Mapping parser and the DoMapping processor.
"""

from collections import defaultdict
import pytest

from vermouth.processors import (
    DoMapping, RepairGraph, CanonicalizeModifications, PDBInput, MakeBonds
)
from vermouth import System
from vermouth import Molecule
from vermouth.map_input import read_mapping_directory
from vermouth.forcefield import ForceField
from vermouth.tests.datafiles import (
    PDB_ALA5,
    PDB_ALA5_CG,
    FF_UNIVERSAL_TEST,
    FF_PEPPLANE,
    FF_MARTINI_TEST,
    MAP_UNIVERSAL_TEST_PEPPLANE
)
from vermouth.tests.helper_functions import equal_graphs


# Pylint doesn't like pytest fixtures
# pylint: disable=redefined-outer-name


@pytest.fixture(scope='session')
def force_fields():
    """
    Read a bunch of force fields fit for testing
    """
    ffs = {}
    ffs['universal-test'] = ForceField(FF_UNIVERSAL_TEST)
    ffs['pepplane'] = ForceField(FF_PEPPLANE)
    ffs['martini-test'] = ForceField(FF_MARTINI_TEST)
    assert set(ffs.keys()) == {'universal-test', 'pepplane', 'martini-test'}
    assert set(ffs['universal-test'].modifications.keys()) >= {'C-ter', 'N-ter'}
    assert set(ffs['universal-test'].blocks.keys()) >= {'ALA',}
    return ffs


@pytest.fixture(scope='session')
def mappings(force_fields):
    """
    Read a bunch of mappings fit for testing
    """
    maps = read_mapping_directory(MAP_UNIVERSAL_TEST_PEPPLANE, force_fields)
    assert set(maps.keys()) == {'universal-test',}
    assert set(maps['universal-test'].keys()) == {'pepplane', 'martini-test'}
    assert set(maps['universal-test']['pepplane'].keys()) == {('AA1', 'AA2'), ('C-ter',), ('N-ter',)}
    return maps


@pytest.fixture()
def ala5_aa(force_fields):
    """
    Read an atomistic alanine oligomer from a PDB file
    """
    system = System()
    PDBInput(PDB_ALA5).run_system(system)
    system.force_field = force_fields['universal-test']
    MakeBonds().run_system(system)
    RepairGraph().run_system(system)
    CanonicalizeModifications().run_system(system)
    return system


@pytest.fixture()
def ala5_cg(force_fields):
    """
    Read a CG alanine oligomer from a PDB file
    """
    system = System()
    PDBInput(PDB_ALA5_CG).run_system(system)
    system.force_field = force_fields['pepplane']
    return system


@pytest.fixture()
def ser3_aa(force_fields):
    """
    Provide a serine oligomer without coordinates
    """
    mol = Molecule(force_field=force_fields['universal-test'])
    ser_block = force_fields['universal-test'].blocks['SER']
    for _ in range(3):
        mol.merge_molecule(ser_block)
    mol.add_edges_from([(10, 12), (21, 23)])
    sys = System(force_field=force_fields['universal-test'])
    assert sys.force_field == force_fields['universal-test']
    sys.add_molecule(mol)
    return sys


def test_pepplane_mapping(force_fields, mappings, ala5_aa, ala5_cg):
    """
    Test a mapping that crosses residue boundaries and is not based on blocks.
    """
    ala5_aa = ala5_aa.copy()
    ala5_cg = ala5_cg.copy()
    DoMapping(mappings, force_fields['pepplane'], attribute_keep=('chain', 'resid'),
              attribute_must=('resname',)).run_system(ala5_aa)

    assert ala5_aa.force_field == ala5_cg.force_field
    assert len(ala5_aa.molecules) == len(ala5_cg.molecules) == 1
    aa_mol = ala5_aa.molecules[0]
    cg_mol = ala5_cg.molecules[0]
    attrnames = ('atomname', 'resname', 'resid')
    for attrname in attrnames:
        print(aa_mol.nodes(data=attrname))
        print(cg_mol.nodes(data=attrname))
    print(aa_mol.edges)
    print(cg_mol.edges)
    assert equal_graphs(aa_mol, cg_mol, node_attrs=attrnames)


def test_spawned_atoms(force_fields, ser3_aa, mappings):
    """
    Make sure that nodes that are not explicitly mapped to are still added, and
    have the right mapping_weights
    """
    DoMapping(mappings, force_fields['martini-test'],
              attribute_keep=('chain', 'resid'),
              attribute_must=('resname',)).run_system(ser3_aa)
    found = ser3_aa.molecules[0]
    assert found.force_field == force_fields['martini-test']
    assert len(found) == 6
    mapped_atoms_per_resid = defaultdict(set)
    for n_idx in found.nodes:
        node = found.nodes[n_idx]
        if node['atomname'] == 'BB':
            mapped_atoms_per_resid[node['resid']].update(node['mapping_weights'].keys())
    for n_idx in found.nodes:
        node = found.nodes[n_idx]
        if node['atomname'] == 'SC1':
            assert node['mapping_weights'] == {idx: 0 for idx in
                                               mapped_atoms_per_resid[node['resid']]}
